/////////////////////////////////////////////////////////////////
//                                                             //
//    ::::::::   ::::::::  :::    ::: ::::::::::: ::::::::::   //
//   :+:    :+: :+:    :+: :+:    :+:     :+:     :+:          //
//         +:+  +:+        +:+    +:+     +:+     +:+          //
//       +#+    +#+        +#+    +:+     +#+     +#++:++#     //
//     +#+      +#+        +#+    +#+     +#+     +#+          //
//    #+#       #+#    #+# #+#    #+#     #+#     #+#          //
//   ##########  ########   ########      ###     ##########   //
//                                                             //
//// needs more optimizating ////////////////////////////////////

import * as PIXI from 'pixi.js';

/** A 2-dimensional point. */
export type X2 = {
   /** The X-value. */
   x: number;
   /** The Y-value. */
   y: number;
};

/** Properties for the `XAnchoredObject` class. */
export type XAnchoredObjectProperties = XObjectProperties & XProperties<XAnchoredObject, 'anchor'>;

/** The data structure of an Aseprite sprite sheet exported with the settings outlined in the `XAnimationResources` type description. */
export type XAnimationData<A extends string = string> = {
   /** Information pertaining to the animation's frames. */
   frames: {
      /** The frame's duration. */
      duration: number;
      /** The frame's dimensions. */
      frame: XArea;
   }[];
   /** The animation's metadata. */
   meta: {
      /** The animation's frame tags. These tags can be added within Aseprite by right-clicking on a frame in the timeline and selecting `New Tag` */
      frameTags: {
         /** The tag name. */
         name: A;
         /** The index of the first frame this tag appears on. */
         from: number;
         /** The index of the last frame this tag appears on. */
         to: number;
      }[];
   };
};

/** Properties for the `XAnimation` class. */
export type XAnimationProperties = XNot<XSpriteProperties, 'crop' | 'frames'> &
   XProperties<XAnimation, 'extrapolate' | 'resources' | 'subcrop'>;

/**
 * An animation's resources, consisting of an image containing each frame, and an instance of animation data used to discern those frames and their properties from said image. These can be exported from Aseprite by using the `Export Sprite Sheet` option from the `File` menu. Make sure to export both `Output File` and `JSON Data`, use the `Array` format, and include `Tags` in the metadata.
 * @example
 * const resources: XAnimationResources = X.inventory(
 *    X.dataAsset('animation-json-data.json'),
 *    X.imageAsset('animation-output-file.png')
 * );
 */
export type XAnimationResources = XInventory<[XData<XAnimationData>, XImage]>;

/** An area with a base position, width, and height. */
export type XArea = X2 & {
   /** The width of the area. */
   w: number;
   /** The height of the area. */
   h: number;
};

/** Properties for the `XAtlas` class. */
export type XAtlasProperties<A extends string = string> = {
   /** Defines the navigators assigned to an atlas. Refer to the `XAtlas` class for a detailed breakdown on the behavior and usage of atlases and navigators. */
   navigators?: XKeyed<XNavigator<A>, A> | void;
} | void;

/** An audio asset, containing an `AudioBuffer`. */
export type XAudio = XAsset<AudioBuffer>;

/** An object that can be safely serialized to JSON. */
export type XBasic = { [k: string]: XBasic } | XBasic[] | string | number | boolean | null | undefined | void;

/** A cardinal direction. One of `down`, `left`, `right`, or `up`. */
export type XCardinal = 'down' | 'left' | 'right' | 'up';

/** A character preset, containing a set of walk sprites and talk sprites in all four cardinal directions. */
export type XCharacterPreset = XKeyed<XKeyed<XSprite, XCardinal>, 'walk' | 'talk'>;

/** Properties for the `XCharacter` class. */
export type XCharacterProperties = XNot<XEntityProperties, 'sprites'> & {
   /** Controls the key associated with a character. This key is arbitrary, but can be used to easily identify the character's object when compared to other objects. In many cases, you will want to animate a specific character on-screen during dialogue. Ideally, that's what this is for. */
   key: string;
   /** Controls the preset a character will use for walk sprites and talk sprites. Refer to the `XCharacter` class for more information. */
   preset: XCharacterPreset;
};

/** Color data, in the form of `[R, G, B, A]`. Each value sits in a range of `0-255`. */
export type XColor = [number, number, number, number];

/**
 * Controls how a sprite is cropped. A crop consists of four properties, one for each side, or edge, of the cropping rectangle. When a value is positive, the corresponding edge will be moved inward by said value. When a value is negative, the corresponding edge will be moved inward from the opposing edge on the sprite's boundaries by said value.
 * @example
 * // crop 5 rows/columns of pixels from each side
 * const sprite = new XSprite({
 *    crop: { top: 5, left: 5, right: 5, bottom: 5 }
 * });
 *
 * // crop a sprite using an X-value, Y-value, width, and height
 * // note: for cartesian alignment, swap the values of "top" and "bottom"
 * const x = 5;
 * const y = 5;
 * const w = 20;
 * const h = 20;
 * const sprite = new XSprite({
 *    crop: { top: y, left: x, right: -x - w, bottom: -y - h }
 * });
 *
 * // if the difference between two opposing crop values is negative, the results are less intuitive; however, they are still predictable (in this example, the crop begins at Y=10 and ends at Y=5, effectively resulting in a reversed crop area)
 * const sprite = new XSprite({
 *    crop: { top: 10, bottom: -5 }
 * });
 */
export type XCrop = XKeyed<number, XSide>;

/** A data asset, containing JSON data. */
export type XData<A extends XBasic = XBasic> = XAsset<A>;

/** Excludes `null`, `undefined`, and `void` from the given type. Internal use only. */
export type XDefined<A> = Exclude<A, null | undefined | void>;

/** Properties for the `XEntity` class. */
export type XEntityProperties = XNot<XObjectProperties, 'objects'> &
   XProperties<XEntity, 'anchor' | 'size' | 'step'> & { sprites?: Partial<XKeyed<XSprite, XCardinal>> | void };

/** A bitmap font. */
export type XFont = { glyphs: XKeyed<XGlyph>; shift: X2; size: number };

/** A list of frames to display. */
export type XFrames = (XImage | null)[];

/** Data representing a bitmap font. */
export type XFontData = { name: string; glyphs: XGlyphData[]; shift: X2; size: number };

/** A glyph, belonging to a bitmap font. */
export type XGlyph = { image: XImage; margin: number; metrics: XArea };

/** Data representing a glyph, belonging to a bitmap font. */
export type XGlyphData = { code: string; margin: number; metrics: XArea; source: string };

/** An `XHost` event handler. Refer to the `XHost` class for guidance in regards to creating and using event handlers. */
export type XHandler<A extends any, B extends any[] = any> =
   | XListener<A, B>
   | {
        /** The listener function. */
        listener: XListener<A, B>;
        /** The listener priority, which controls the execution order of this listener. */
        priority: number;
     };

/** An image asset, containing a `PIXI.Texture`. */
export type XImage = XAsset<PIXI.Texture>;

/** Properties for the `XInput` class. */
export type XInputProperties = {
   /** The keycodes to listen for. */
   codes?: string[];
   /** The element to attach event listeners to. Defaults to `window`. */
   target?: HTMLElement;
} | void;

/** An audio instance. Refer to the `XDaemon` type for a more in-depth explanation of how instances are created and used. */
export type XInstance = {
   /** The instance's `AudioContext`. */
   context: AudioContext;
   /** The instance's parent daemon. */
   daemon: XDaemon;
   /** Controls whether this instance should loop playback or not. */
   loop: boolean;
   /** Controls the instance's gain. */
   gain: AudioParam;
   /** Controls the instance's sampling rate. This alters both the speed and pitch of the audio. */
   rate: AudioParam;
   /** Stops playback, rendering this instance inert. */
   stop(): void;
};

/** Groups assets together so they can be batch-loaded and batch-unloaded. */
export type XInventory<A extends XAsset[] = XAsset[]> = XAsset<A>;

/** An object with strictly-defined keys. */
export type XKeyed<A = any, B extends string = any> = { [x in B]: A };

/** An `XHandler` listener function. */
export type XListener<A extends any, B extends any[]> = (
   this: A,
   /** The data passed to the listener. */
   ...data: B
) => any;

/** A valid navigator key, used in `XAtlas` navigation functions. */
export type XNavigatorKey<A extends string = string> = A | null | void;

/** Properties for the `XNavigator` class. */
export type XNavigatorProperties<A extends string = string> = XProperties<XNavigator<A>, 'objects' | 'position'> &
   ({
      /** A provider which controls whether or not the X and Y axes of a navigator should be flipped. */
      flip?: XNavigator<A>['flip'] | void;
      /** A provider which controls what elements are available in a navigator's grid. */
      grid?: XNavigator<A>['grid'] | void;
      /** A provider which controls what a navigator's "next" target is. When specified as a function, this property can be used to perform additional actions upon navigation, such as playing a sound effect or altering a displayed value. */
      next?: XNavigator<A>['next'] | void;
      /** A provider which controls what a navigator's "prev" target is. Much like the `next` property, this property can be used to perform additional actions upon navigation when specified as a function. */
      prev?: XNavigator<A>['prev'] | void;
   } | void);

/** Excludes keys from the given type. Internal use only. */
export type XNot<A, B extends string> = { [x in Exclude<keyof XDefined<A>, B>]?: XDefined<A>[x] };

/** The parent type for both objects and rendering layers. Contains a PIXI container and child objects. */
export type XObjectParent = { container: PIXI.Container; objects: XObject[] };

/** Properties for the `XObject` class. */
export type XObjectProperties = XProperties<
   XObject,
   | 'alpha'
   | 'blend'
   | 'border'
   | 'container'
   | 'fill'
   | 'font'
   | 'friction'
   | 'gravity'
   | 'metadata'
   | 'objects'
   | 'parallax'
   | 'position'
   | 'priority'
   | 'rotation'
   | 'scale'
   | 'spin'
   | 'stroke'
   | 'tint'
   | 'velocity'
> &
   ({
      /** The PIXI filters to apply to this object. */
      filters?: PIXI.Filter[] | void;
   } | void);

/** Recursively converts types into primitive forms where possible. Internal use only. */
export type XPrimitive<A> = A extends string | number | boolean | null | undefined
   ? A | void
   : //
   A extends { value: number }
   ? number | void
   : //
   A extends HTMLElement
   ? HTMLElement
   : //
   A extends PIXI.Container
   ? PIXI.Container
   : //
   A extends XAnimationResources
   ? XAnimationResources | void
   : //
   A extends XAsset<infer B>
   ? XAsset<B> | void
   : //
   A extends XColor
   ? XColor | void
   : //
   A extends XFrames
   ? XFrames | void
   : //
   A extends XObject[]
   ? XObject[] | void
   : //
   A extends XRendererLayerModifier[]
   ? XRendererLayerModifier[] | void
   : //
   A extends XVector
   ? { x?: number | void; y?: number | void } | number | void
   : //
   A extends (infer B)[]
   ? XPrimitive<B>[] | void
   : //
   A extends XKeyed
   ? Partial<{ [x in keyof A]?: XPrimitive<A[x]> }> | void
   : //
     never;

/** Generates a "properties" object from a given type. Internal use only. */
export type XProperties<A extends XKeyed, B extends keyof A> = XPrimitive<{ [x in B]: A[x] }>;

/**
 * Converts the input type into an output type consisting of a union between the input type and a function returning the output type. Best paired with `X.provide`. I know, it's confusing.
 * @example
 * // delcare a function with a provider as an argument
 * function example1 (value: XProvider<string>) {
 *    console.log(X.provide(value));
 * }
 *
 * // call that function!
 * example1('Howdy!'); // logs "Howdy!"
 * example1(() => 'Howdy!'); // logs "Howdy!"
 * example1(() => () => 'Howdy!'); // logs "Howdy!"
 * example1(() => () => () => 'Howdy!'); // logs "Howdy!" alright you get the point-
 *
 * // you can also declare arguments that will be passed into the provider, given that it's a function
 * function example2 (value: XProvider<number, [number, number, number]>) {
 *    console.log(X.provide(value, 10, 20, 30, 10, 20, 30, 10, 20, 30));
 * }
 *
 * // call that function!
 * example2((rand1, rand2, rand3) => rand1 + rand2 + rand3); // logs 60
 * example2((rand1, rand2, rand3) => rand1 * rand2 * rand3); // logs 6000
 */
export type XProvider<A, B extends any[] = []> = A | ((...args: B) => XProvider<A, B>);

/** A region between two points, in the form of `[{ x: minX, y: minY }, { x: maxX, y: maxY }]`. */
export type XRegion = [X2, X2];

/** A renderer layer modifier. These modifiers can be added and removed from a rendering layer's `modifiers` array to alter the behavior of said layer. A value of `"ambient"` prevents a layer from updating unless its renderer's camera, scale, canvas size, or quality value changes. This saves on performance where static backgrounds are involved. A value of `"static"` causes a layer to ignore its renderer's camera position. This ensures that objects on said layer won't move with the camera, which is useful for menus. A value of `"vertical"` will cause a layer to treat the `position.y` of its objects as those objects' default `priority` value. In effect, objects, when rendered, will be automatically sorted by their Y-position unless those objects specify their own `priority` values. A value of `"secure"` means the layer will not be impacted by the shake effect.*/
export type XRendererLayerModifier = 'ambient' | 'static' | 'vertical' | 'secure';

/** Properties for the `XRendererLayer` class. */
export type XRendererLayerProperties = XProperties<XRendererLayer, 'active' | 'modifiers'> &
   ({
      /** The PIXI filters to apply to this object. */
      filters?: PIXI.Filter[] | void;
   } | void);

/** Properties for the `XRenderer` class. */
export type XRendererProperties<A extends string = string> = XProperties<
   XRenderer<A>,
   'alpha' | 'camera' | 'region' | 'rotation' | 'shake' | 'zoom'
> &
   ({
      /** If this value is an HTML element, it will be used as the container for the renderer, but if the value is a string, it will treat it as a query selector. The container of a renderer is the HTML element where all of the renderer's canvases are added. */
      container?: HTMLElement | string | void;
      /** Controls what layers this renderer will utilize. The keys of this property determine the renderer's layer keys, and the values correspond to the modifiers each layer should be initialized with. */
      layers?: XKeyed<XRendererLayer, A> | void;
      /** Renderer quality. Multiplies the coordinate scale. */
      quality?: number | void;
      /** Renderer size. */
      size?: XPrimitive<XVector>;
   } | void);

/** Extrapolates a promise result. */
export type XResult<A> = A extends () => Promise<infer B> ? B : never;

/** An audio router, used to customize how audio is routed in a daemon. By default, a daemon will simply connect the `input` node to the `context.destination` node. A custom router allows you to change the way audio is routed between these two points. Note that, if you don't in some way connect `input` to `context.destination`, the audio will not reach the output device. */
export type XRouter = (context: AudioContext, input: GainNode) => void;

/** A pixel shader, used to transform an image's pixel data. Each pixel in the image is passed through one at a time, in the form of `[color, position]`. The image's  width and height is also passed in as a 2-dimensional point. */
export type XShader = (
   /** The color of the pixel. */
   color: XColor,
   /** The position of the pixel. */
   position: X2,
   /** The size of the image, with `size.x` referring to width and `size.y` referring to height. */
   size: X2
) => XColor;

/** A rectangular side. One of `bottom`, `left`, `right`, or `top`. */
export type XSide = 'bottom' | 'left' | 'right' | 'top';

/** Properties for the `XSizedObject` class. */
export type XSizedObjectProperties = XAnchoredObjectProperties & XProperties<XSizedObject, 'size'>;

/** Properties for the `XSprite` class. */
export type XSpriteProperties = XAnchoredObjectProperties &
   XProperties<XSprite, 'active' | 'crop' | 'duration' | 'frames' | 'increment' | 'index' | 'offset'>;

/** An object's computed style at render time. Internal use only. */
export type XStyle = {
   blend: GlobalCompositeOperation;
   border: number;
   fill: string;
   font: string;
   stroke: string;
   tint: number;
};

/** Properties for the `XText` class. */
export type XTextProperties = XAnchoredObjectProperties &
   XProperties<XText, 'charset' | 'content' | 'plain' | 'spacing'>;

/** A transform, in the form of `[{ x: positionX, y: positionY }, rotation, { x: scaleX, y: scaleY }]`. Internal use only. */
export type XTransform = [XVector, number, XVector];

/** An asset, consisting of a loader, unloader, and source string for clarity in errors. */
export class XAsset<A = any> {
   /** This asset's loader method. Returns a promise which resolves with the loaded value. */
   loader: () => Promise<A>;
   /** This asset's source string. Used in thrown errors to help identify the faulty asset. */
   source: string;
   /** The state of the asset, containing the current loaded value, if any. */
   state = { retainers: [] as XAsset[], value: void 0 as A | void };
   /** This asset's unloader, which unassigns the loaded value and returns a promise which resolves when the asset is fully unloaded. Certain assets, such as images and audios, use this function to perform necessary cleanup operations. */
   unloader: () => Promise<void>;
   /** Gets the value if defined, or throws an error stating the asset must be loaded. If you prefer not to recieve an error, check if `state.value` is defined before accessing this property. */
   get value () {
      const value = this.state.value;
      if (value === void 0) {
         throw new ReferenceError(`The asset of "${this.source}" is not currently loaded!`);
      } else {
         return value;
      }
   }
   constructor ({
      loader,
      source,
      unloader
   }: {
      /** This asset's loader method. Returns a promise which resolves with the loaded value. */
      loader: () => Promise<A>;
      /** This asset's source string. Used in thrown errors to help identify the faulty asset. */
      source: string;
      /** This asset's unloader, which unassigns the loaded value and returns a promise which resolves when the asset is fully unloaded. Certain assets, such as images and audios, use this function to perform necessary cleanup operations. */
      unloader: () => Promise<void>;
   }) {
      this.loader = loader;
      this.source = source;
      this.unloader = unloader;
   }
   /** Requests for the asset to be loaded. */
   async load (
      /** The retainer to register. This will act as a secondary source of keeping the asset loaded. */
      retainer: XAsset = this
   ) {
      this.state.retainers.includes(retainer) || this.state.retainers.push(retainer);
      if (this.state.value === void 0) {
         this.state.value = await this.loader();
      }
   }
   /** Requests for the asset to be unloaded. */
   async unload (
      /** The retainer to unregister. All previously registered retainers must be passed through here to unload the asset. */
      retainer: XAsset = this
   ) {
      this.state.retainers.includes(retainer) && this.state.retainers.splice(this.state.retainers.indexOf(retainer), 1);
      if (this.state.value !== void 0 && this.state.retainers.length === 0) {
         this.state.value = await this.unloader();
      }
   }
}

/**
 * An "atlas" used to manage `XNavigator` objects. Each atlas consists of a map of navigator objects, which can be traversed and managed via the `XAtlas` methods. The `"to"`, `"from"`, and `"move"` events on each navigator allow for automatic handling.
 * @example
 * // example renderer
 * const renderer = new XRenderer({ layers: { atlasLayer: [ 'static' ] } });
 *
 * // create an atlas with navigators "a", "b", and "c"
 * const atlas = new XAtlas<'a' | 'b' | 'c'>({
 *    navigators: {
 *       a: new XNavigator({
 *          next: 'b',
 *          prev: null
 *       }),
 *       b: new XNavigator({
 *          next: 'c',
 *          prev: 'a'
 *       }),
 *       c: new XNavigator({
 *          prev: 'b'
 *       })
 *    }
 * });
 *
 * // add handlers for attaching and detaching
 * atlas.navigators.a.on('from', () => {
 *    atlas.attach(renderer, 'atlasLayer', );
 * });
 */
export class XAtlas<A extends string = string> {
   /** The atlas's assigned navigators. */
   navigators: XKeyed<XNavigator<A>, A>;
   /** The atlas's state, containing the currently active navigator. */
   state = { navigator: null as A | null };
   constructor ({ navigators = {} as XKeyed<XNavigator<A>, A> }: XAtlasProperties<A> = {}) {
      this.navigators = navigators;
   }
   /** Attaches objects to a navigator or navigators to the given renderer. */
   attach<B extends XRenderer<any>> (
      /** The renderer to attach the objects to. */
      renderer: B,
      /** The key of the rendering layer the objects should be added to. */
      layer: B extends XRenderer<infer C> ? C : never,
      /** The keys of the navigators whose objects should be attached. */
      ...navigators: A[]
   ) {
      for (const nav of navigators) {
         nav in this.navigators && this.navigators[nav].attach(renderer, layer);
      }
   }
   /** Detaches objects from a navigator or navigators to the given renderer. */
   detach<B extends XRenderer<any>> (
      /** The renderer to detach the objects from. */
      renderer: B,
      /** The key of the rendering layer the objects should be removed from. */
      layer: B extends XRenderer<infer C> ? C : never,
      /** The keys of the navigators whose objects should be attached. */
      ...navigators: A[]
   ) {
      for (const nav of navigators) {
         nav in this.navigators && this.navigators[nav].detach(renderer, layer);
      }
   }
   /** Returns the current navigator, or `null` if one is not currently active. */
   navigator () {
      return this.state.navigator === null ? null : this.navigators[this.state.navigator];
   }
   /** Attempts to move the current navigator's position by the specified offset. If the resulting position is outside of the range of the current navigator's grid, the value will be wrapped around. If the current navigator's computed selection before and after the seeking occurs differs, that navigator's `"move"` event will be fired with both the old and new selection value. */
   seek ({ x = 0, y = 0 }: Partial<X2> | void = {}) {
      const nav = this.navigator();
      if (nav) {
         const origin = nav.selection();
         const row = X.provide(nav.grid, nav, this);
         const flip = X.provide(nav.flip, nav, this);
         nav.position.x = Math.min(Math.max(nav.position.x, 0), row.length - 1);
         nav.position.x += flip ? y : x;
         if (row.length - 1 < nav.position.x) {
            nav.position.x = 0;
         } else if (nav.position.x < 0) {
            nav.position.x = row.length - 1;
         }
         const column = row[nav.position.x] ?? [];
         nav.position.y = Math.min(Math.max(nav.position.y, 0), column.length - 1);
         nav.position.y += flip ? x : y;
         if (column.length - 1 < nav.position.y) {
            nav.position.y = 0;
         } else if (nav.position.y < 0) {
            nav.position.y = column.length - 1;
         }
         const selection = nav.selection();
         origin === selection || nav.fire('move', this, nav, origin, selection);
      }
   }
   /**
    * Triggers a navigation in the atlas. Based on the `action` specified, the relevant provider in the current navigator (if set) will be computed. The result of this computation is then passed into this atlas's `switch` method.
    * @example
    * const atlas = new XAtlas({
    *    // navigator "a"
    *    a: new XNavigator({
    *       // "next" provider on navigator "a"
    *       next () {
    *          return Math.random() < 0.5 ? 'b' : null;
    *       }
    *    }),
    *    // navigator "b"
    *    b: new XNavigator({
    *       prev: 'a'
    *    })
    * });
    *
    * atlas.switch('a'); // current navigator is now "a"
    * atlas.navigate('next'); // the "next" provider on navigator "a" is called
    *
    * // 50% chance of being true, as made apparent by the code in the "next" provider on navigator "a"
    * if (atlas.state.navigator === 'b') {
    *    console.log("We're in B now!\nThat's B for business!");
    *
    *    // now that we're in navigator "b" we can return back to navigator "a" since the "prev" provider on navigator "b" is just a fixed value of "a"
    *    atlas.navigate('prev'); // current navigator is now "a" again
    *
    *    // now we get to do it all over again!
    *    atlas.navigate('next'); // the "next" provider on navigator "a" is called yet again. who knows what'll happen now!
    * } else {
    *    console.log("Luck is not on our side today.");
    * }
    */
   navigate (action: 'next' | 'prev') {
      const nav = this.navigator();
      if (nav) {
         const result = X.provide(nav[action], nav, this) as XNavigatorKey<A>;
         result === void 0 || this.switch(result);
      }
   }
   /** Sets the currently active navigator to the navigator with the given key. If `null` is specified, the currently active navigator is un-set. The `"to"` event in the current navigator (if set) as well as the `"from"` event in the target navigator (again, if set) will both be fired. */
   switch (name: A | null) {
      if (name !== void 0) {
         let next: XNavigator<A> | null = null;
         if (typeof name === 'string') {
            if (name in this.navigators) {
               next = this.navigators[name];
            } else {
               return;
            }
         }
         const nav = this.navigator();
         nav && nav.fire('to', this, name, nav);
         next && next.fire('from', this, this.state.navigator, nav);
         this.state.navigator = name;
      }
   }
}

/** A map-based cache. Values are computed once per key via the `compute` function passed to the constructor. */
export class XCache<A, B> extends Map<A, B> {
   /** The function used to compute a new value. */
   compute: (
      /** The key object. */
      key: A
   ) => B;
   constructor (
      /** The function used to compute a new value. */
      compute: (
         /** The key object. */
         key: A
      ) => B
   ) {
      super();
      this.compute = compute;
   }
   /** Retrieves a value in the cache. If no value is present, the computation function is used to generate one. */
   of<C = B> (key: A) {
      this.has(key) || this.set(key, this.compute(key));
      return this.get(key) as any as C;
   }
}

/**
 * An audio daemon, used to create and manage audio instances. An audio daemon sets defualt values for `gain`, `loop`, and `rate`, and any instances created from said daemon will inherit those values by default. Once an instance is created, its `gain`, `loop`, and `rate` properties can be modified independently.
 * @example
 * // create an audio asset
 * const audio = X.audioAsset('path/to/file.wav');
 *
 * // assuming you are in an async context, await the asset's "load" method
 * await audio.load();
 *
 * // create a daemon, with a default rate of `0.5` and a default loop setting of `true`.
 * const daemon = X.daemon(audio, { gain: 0.5, loop: true });
 *
 * // create (and thus, play) the instance (it will be at half speed/pitch, and loop itself)
 * const instance = daemon.instance();
 *
 * // just for fun, ramp the instance's rate to 2 over a period of 1 second (1,000 milliseconds)
 * await instance.gain.modulate(1000, 2);
 *
 * // then, when you're bored, you can stop it like so
 * instance.stop();
 *
 * // if there are no running instances, the asset can be safely unloaded
 * audio.unload();
 */
export class XDaemon {
   private $daemon = { instances: [] as XInstance[] };
   /** The daemon's audio asset. */
   audio: XAudio;
   /** The `AudioContext` associated with this daemon. */
   context: AudioContext;
   /** The default `gain` value for all created instances. */
   gain: number;
   /** The default `loop` value for all created instances. */
   loop: boolean;
   /** The default `rate` value for all created instances. */
   rate: number;
   /** This daemon's audio router. Refer to the `XRouter` type for information about audio routers. */
   router: XRouter;
   /** The instances stored within this daemon. */
   get instances () {
      return this.$daemon.instances;
   }
   constructor (
      audio: XAudio,
      {
         /** The `AudioContext` associated with this daemon. */
         context = new AudioContext(),
         /** The default `gain` value for all created instances. */
         gain = 1,
         /** The default `loop` value for all created instances. */
         loop = false,
         /** The default `rate` value for all created instances. */
         rate = 1,
         /** This daemon's audio router. Refer to the `XRouter` type for information about audio routers. */
         router = ((context: AudioContext, input: GainNode) => input.connect(context.destination)) as XRouter
      } = {}
   ) {
      this.audio = audio;
      this.context = context;
      this.gain = gain;
      this.loop = loop;
      this.rate = rate;
      this.router = router;
   }
   /** Creates an instance from this daemon. */
   instance (
      /** The start time in seconds. Defaults to `0`. */
      offset = 0,
      /** Whether or not this instance should be stored in the parent daemon. Defaults to `false`. */
      store = false
   ) {
      const gain = this.context.createGain();
      gain.gain.value = this.gain;
      this.router(this.context, gain);
      const source = this.context.createBufferSource();
      if (this.audio.state.value) {
         source.buffer = this.audio.value;
      } else {
         X.status(`Error: Cannot invoke the unloaded asset of "${this.audio.source}"`, {
            backgroundColor: '#f00',
            color: '#fff',
            fontSize: '12px'
         });
      }
      source.loop = this.loop;
      source.playbackRate.value = this.rate;
      source.connect(gain);
      const instance = {
         context: this.context,
         daemon: this,
         gain: gain.gain,
         get loop () {
            return source.loop;
         },
         set loop (value) {
            source.loop = value;
         },
         rate: source.playbackRate,
         stop: () => {
            if (source.buffer) {
               store && this.instances.splice(this.instances.indexOf(instance), 1);
               source.stop();
               source.disconnect(gain);
               gain.disconnect();
               source.buffer = null;
            }
         }
      };
      store && this.instances.push(instance);
      if (source.buffer) {
         source.addEventListener('ended', () => {
            instance.loop || instance.stop();
         });
         source.start(0, offset);
      }
      return instance;
   }
}

/** An audio effect with a dry `throughput` channel and wet `input` and `output` channels. */
export class XEffect {
   /** The effect's `AudioContext`. */
   context: AudioContext;
   /** The `AudioNode` used to generate the effect. */
   input: AudioNode;
   /** A gain control for the wet, effected audio. This value should not be modified directly, instead, use `XEffect.value` to control the mix. */
   output: GainNode;
   /** A throughput for dry, incoming audio. This value should not be modified directly, instead, use `XEffect.value` to control the mix. */
   throughput: GainNode;
   /** Controls the ratio of the mix. When this value is `0`, the output is fully dry, with no effected sound. When this value is `1`, the output is fully wet, with none of the original signal. At `0.5`, there is an equal amount of dry and wet sound. */
   get value () {
      return this.output.gain.value;
   }
   set value (value) {
      this.output.gain.value = value;
      this.throughput.gain.value = 1 - value;
   }
   constructor (
      /** The `AudioContext` to use for this effect. This context must match that of the internal effect node. */
      context: AudioContext,
      /** The internal effect node. This can be a convolver, filter, etc. */
      node: AudioNode,
      /** The initial output value to set. */
      value = 1
   ) {
      this.context = context;
      this.input = node;
      this.output = context.createGain();
      this.throughput = context.createGain();
      this.input.connect(this.output);
      this.value = value;
   }
   /** Connects this effect's output to another effect, audio node, or audio context. */
   connect (target: XEffect | AudioNode | AudioContext) {
      if (target instanceof AudioContext) {
         this.output.connect(target.destination);
         this.throughput.connect(target.destination);
      } else if (target instanceof AudioNode) {
         this.output.connect(target);
         this.throughput.connect(target);
      } else {
         this.output.connect(target.input);
         this.throughput.connect(target.throughput);
      }
   }
   /** Disconnects this effect's output from another previously connected effect, audio node, or audio context. */
   disconnect (target?: XEffect | AudioNode | AudioContext) {
      if (target instanceof AudioContext) {
         this.output.disconnect(target.destination);
         this.throughput.disconnect(target.destination);
      } else if (target instanceof AudioNode) {
         this.output.disconnect(target);
         this.throughput.disconnect(target);
      } else if (target) {
         this.output.disconnect(target.input);
         this.throughput.disconnect(target.throughput);
      } else {
         this.output.disconnect();
      }
   }
   /** Interpolates this effect's value over a period of time. See `XNumber.interpolate` for specifics. */
   interpolate (duration: number, ...points: number[]) {
      return XNumber.prototype.interpolate.call(this, duration, ...points);
   }
   /** Modulates this effect's value over a period of time. See `XNumber.modulate` for specifics. */
   modulate (duration: number, ...points: number[]) {
      return XNumber.prototype.modulate.call(this, duration, ...points);
   }
}

/** An event host, capable of firing one-time and repeating events to listeners with variable priorities. */
export class XHost<A extends XKeyed<any[]> = {}> {
   /** The events stored in the event host. This should not be modified directly. */
   events: { [B in keyof A]?: { handlers: XHandler<any, A[B]>[]; priorities: number } } = {};
   /** Fires an event. */
   fire<B extends keyof A> (
      /** The event's name. */
      name: B,
      /** The data to pass into each listener. */
      ...data: A[B]
   ) {
      const event = this.events[name];
      if (event) {
         return event.handlers.slice().map(handler => {
            return (typeof handler === 'function' ? handler : handler.listener).call(this, ...data);
         });
      } else {
         return [];
      }
   }
   /** Detaches an event handler from an event. */
   off<B extends keyof A> (
      /** The name of the event to detach the handler from. */
      name: B,
      /** The handler to detach. */
      handler: XHandler<this, A[B]>
   ) {
      const event = this.events[name];
      if (event) {
         const index = event.handlers.indexOf(handler);
         if (index > -1) {
            event.handlers.splice(index, 1);
            if (typeof handler === 'object' && handler.priority !== 0) {
               event.priorities--;
            }
         }
      }
      return this;
   }
   /** Attaches an event handler to an event. In this form, no handler is specified, and a promise which resolves when the event is next fired is returned. */
   on<B extends keyof A>(
      /** The name of the event. */
      name: B
   ): Promise<A[B]>;
   /** Attaches an event handler to an event. In this form, no handler is specified, and a promise which resolves when the event is next fired is returned. */
   on<B extends keyof A>(
      /** The name of the event. */
      name: B,
      /** The priority to use for the internal promise-resolving handler. */
      priority: number
   ): Promise<A[B]>;
   /** Attaches an event handler to an event. */
   on<B extends keyof A>(
      /** The name of the event. */
      name: B,
      /** The event handler to attach. */
      listener: XHandler<this, A[B]>
   ): this;
   on<B extends keyof A> (name: B, a2: number | XHandler<this, A[B]> = 0) {
      if (typeof a2 === 'number') {
         return new Promise(resolve => {
            const singleton = {
               listener: (...data: A[B]) => {
                  this.off(name, singleton);
                  resolve(data);
               },
               priority: a2
            };
            this.on(name, singleton);
         });
      } else {
         const event = (this.events[name] ??= { handlers: [] as XHandler<this, A[B]>[], priorities: 0 });
         event.handlers.push(a2);
         if (typeof a2 === 'object' && a2.priority !== 0) {
            event.priorities++;
         }
         if (event.priorities > 0) {
            event.handlers.sort(
               (handler1, handler2) =>
                  (typeof handler1 === 'function' ? 0 : handler1.priority) -
                  (typeof handler2 === 'function' ? 0 : handler2.priority)
            );
         }
         return this;
      }
   }
   onWith<B extends keyof A> (
      /** The name of the event. */
      name: B,
      /** A function which returns an event listener. */
      provider: (
         /** The host object. */
         host: this
      ) => XHandler<this, A[B]>
   ) {
      return this.on(name, provider(this));
   }
}

/** Handles keyboard input from multiple key codes. If a key is held while the window or tab is switched, stuck-key glitches may occur. There are a few potential fixes for this which you may want to consider implementing in your own code when using this class. If you choose to do so, remember that all active keys can be cleared with `state.codes.clear`. */
export class XInput extends XHost<XKeyed<[string], 'down' | 'up'>> {
   /** The input's state, containing all currently active codes. */
   state = { codes: new Set<string>() };
   constructor ({ codes = [], target = window as any }: XInputProperties = {}) {
      super();
      target.addEventListener('keyup', ({ code }) => {
         if (codes.includes(code) && this.state.codes.has(code)) {
            this.state.codes.delete(code);
            this.fire('up', code);
         }
      });
      target.addEventListener('keydown', ({ code }) => {
         if (codes.includes(code) && !this.state.codes.has(code)) {
            this.state.codes.add(code);
            this.fire('down', code);
         }
      });
   }
   /** True if this key is currently active, in other words, if at least one key this input system listens for is pressed. */
   active () {
      return this.state.codes.size > 0;
   }
}

/** An audio mixer, used to chain `XEffect` instances together. */
export class XMixer {
   /** The audio context used by this mixer. */
   context: AudioContext;
   /** The effects present in this mixer's effect chain. */
   effects = new Proxy([] as XEffect[], {
      deleteProperty: (target, key: `${number}`) => {
         typeof key === 'string' && this.update(+key);
         return delete target[key];
      },
      set: (target, key: `${number}`, value) => {
         typeof key === 'string' && this.update(+key, value);
         target[key] = value;
         return true;
      }
   });
   /** The input node. */
   input: GainNode;
   /** The output node. */
   output: GainNode;
   constructor (
      /** The audio context used by this mixer. All effects in the mixer must use this context. */
      context: AudioContext,
      /** The effects present in this mixer's effect chain. */
      effects: XEffect[] = []
   ) {
      this.context = context;
      this.input = context.createGain();
      this.output = context.createGain();
      this.input.connect(this.output);
      this.effects.push(...effects);
   }
   /** Updates a connection in the mixer. Internal use only. */
   update (index: number, value?: XEffect | void) {
      if (Number.isInteger(index) && index > -1 && index < this.effects.length + 1) {
         const prevValue = (this.effects[index - 1] ?? this.input) as XEffect | GainNode;
         const currValue = this.effects[index];
         const nextValue = (this.effects[index + 1] ?? this.output) as XEffect | GainNode;
         if (currValue) {
            if (prevValue instanceof XEffect) {
               prevValue.disconnect(currValue);
            } else {
               prevValue.disconnect(currValue.input);
               prevValue.disconnect(currValue.throughput);
            }
            currValue.disconnect(nextValue);
         } else {
            prevValue.disconnect(nextValue);
         }
         if (value) {
            if (prevValue instanceof XEffect) {
               prevValue.connect(value);
            } else {
               prevValue.connect(value.input);
               prevValue.connect(value.throughput);
            }
            value.connect(nextValue);
         } else {
            if (prevValue instanceof XEffect) {
               prevValue.connect(nextValue);
            } else if (nextValue instanceof XEffect) {
               prevValue.connect(nextValue.input);
               prevValue.connect(nextValue.throughput);
            } else {
               prevValue.connect(nextValue);
            }
         }
      }
   }
}

/** Can be exported from a file and then imported asynchronously and concurrently with `import`. */
export class XModule<A, B extends any[]> {
   /** The display name of the module. */
   name: string;
   /** The script which returns the promise for the module content. */
   script: (...args: B) => Promise<A>;
   /** The cached promise, used by `import` internally. */
   promise = void 0 as Promise<A> | void;
   constructor (
      /** The display name of the module. */
      name: string,
      /** The script which returns the promise for the module content. */ script: (...args: B) => Promise<A>
   ) {
      this.name = name;
      this.script = script;
   }
   /** Imports the module. If already imported before, the same promise is returned. */
   import (...args: B) {
      X.status(`IMPORT MODULE: ${this.name}`, { color: '#07f' });
      return (this.promise ??= new Promise<A>(async resolve => {
         try {
            resolve(await this.script(...args));
            X.status(`MODULE INITIALIZED: ${this.name}`, { color: '#0f0' });
         } catch (error) {
            X.status(`MODULE ERROR: ${this.name}`, { color: '#f00' });
            console.error(error);
         }
      }));
   }
}

export class XNavigator<A extends string = string> extends XHost<
   XKeyed<[XAtlas<A>, XNavigatorKey<A>, XNavigator<A> | null], 'from' | 'to'> & {
      move: [XAtlas<A>, XNavigator<A>, XBasic, XBasic];
   }
> {
   flip: XProvider<boolean, [XNavigator<A>, XAtlas<A>]>;
   grid: XProvider<XBasic[][], [XNavigator<A>, XAtlas<A> | void]>;
   next: XProvider<XNavigatorKey<A>, [XNavigator<A>, XAtlas<A>]>;
   objects: XObject[];
   position: X2;
   prev: XProvider<XNavigatorKey<A>, [XNavigator<A>, XAtlas<A>]>;
   constructor ({
      flip = false,
      grid = [],
      next = void 0,
      objects = [],
      position: { x = 0, y = 0 } = {},
      prev = void 0
   }: XNavigatorProperties<A> = {}) {
      super();
      this.flip = flip;
      this.grid = grid;
      this.next = next;
      this.objects = objects.map(object => (object instanceof XObject ? object : new XObject(object)));
      this.position = { x, y };
      this.prev = prev;
   }
   attach<B extends XRenderer<any>> (renderer: B, layer: B extends XRenderer<infer C> ? C : never) {
      renderer.attach(layer, ...this.objects);
   }
   detach<B extends XRenderer<any>> (renderer: B, layer: B extends XRenderer<infer C> ? C : never) {
      renderer.detach(layer, ...this.objects);
   }
   selection () {
      return (X.provide(this.grid, this)[this.position.x] ?? [])[this.position.y];
   }
}

export class XNumber {
   value: number;
   constructor (value = 0) {
      this.value = value;
   }
   interpolate (duration: number, ...points: number[]) {
      const base = X.timer.value;
      const value = this.value;
      let active = true;
      X.cache.tasks.get(this)?.();
      X.cache.tasks.set(this, () => (active = false));
      return X.when(() => {
         if (active) {
            const elapsed = X.timer.value - base;
            if (elapsed < duration) {
               this.value = X.math.linear(elapsed / duration, value, ...points);
               return false;
            } else {
               X.cache.tasks.delete(this);
               this.value = points.length > 0 ? points[points.length - 1] : value;
            }
         }
         return true;
      });
   }
   modulate (duration: number, ...points: number[]) {
      const base = X.timer.value;
      const value = this.value;
      let active = true;
      X.cache.tasks.get(this)?.();
      X.cache.tasks.set(this, () => (active = false));
      return X.when(() => {
         if (active) {
            const elapsed = X.timer.value - base;
            if (elapsed < duration) {
               this.value = X.math.bezier(elapsed / duration, value, ...points);
               return false;
            } else {
               X.cache.tasks.delete(this);
               this.value = points.length > 0 ? points[points.length - 1] : value;
            }
         }
         return true;
      });
   }
   async step (speed: number, value = this.value) {
      await this.modulate((Math.abs(this.value - value) / speed) * 1000, value);
   }
}

/** The parent class for all rendered objects. */
export class XObject extends XHost<{ render: []; tick: [] }> implements XObjectParent {
   private $object = {
      container: new PIXI.Container(),
      dirty: false,
      objects: new Proxy([] as XObject[], {
         set: (target, key: `${number}`, newValue) => {
            this.$object.dirty = true;
            target[key] = newValue;
            return true;
         },
         deleteProperty: (target, key: `${number}`) => {
            this.$object.dirty = true;
            delete target[key];
            return true;
         }
      })
   };
   alpha = new XNumber();
   blend: GlobalCompositeOperation | void;
   border: number | void;
   container = new PIXI.Container();
   fill: string | void;
   font: string | void;
   friction = new XNumber();
   gravity = { angle: new XNumber(), extent: new XNumber() };
   metadata: XKeyed<XBasic>;
   parallax = new XVector();
   position = new XVector();
   priority = new XNumber();
   rotation = new XNumber();
   scale = new XVector();
   spin = new XNumber();
   stroke: string | void;
   tint: number | void;
   velocity = new XVector();
   get objects () {
      return this.$object.objects;
   }
   set objects (value) {
      this.$object.objects.splice(0, this.$object.objects.length);
      this.$object.objects.push(...value);
   }
   constructor ({
      alpha = 1,
      blend = void 0,
      border = void 0,
      fill = void 0,
      font = void 0,
      filters = [],
      friction = 1,
      gravity: { angle = 0, extent = 0 } = {},
      metadata = {},
      objects = [],
      parallax = 0,
      position = 0,
      priority = 0,
      rotation = 0,
      scale = 1,
      spin = 0,
      stroke = void 0,
      tint = void 0,
      velocity = 0
   }: XObjectProperties = {}) {
      super();
      this.alpha.value = alpha;
      this.blend = blend;
      this.border = border;
      this.fill = fill;
      this.font = font;
      this.friction.value = friction;
      this.gravity.angle.value = angle;
      this.gravity.extent.value = extent;
      this.metadata = metadata;
      this.objects.push(...objects);
      if (typeof parallax === 'number') {
         this.parallax.inherit(parallax);
      } else {
         this.parallax.inherit(parallax.x ?? 0, parallax.y ?? 0);
      }
      if (typeof position === 'number') {
         this.position.inherit(position);
      } else {
         this.position.inherit(position.x ?? 0, position.y ?? 0);
      }
      this.priority.value = priority;
      this.rotation.value = rotation;
      if (typeof scale === 'number') {
         this.scale.inherit(scale);
      } else {
         this.scale.inherit(scale.x ?? 1, scale.y ?? 1);
      }
      this.spin.value = spin;
      this.stroke = stroke;
      this.tint = tint;
      if (typeof velocity === 'number') {
         this.velocity.inherit(velocity);
      } else {
         this.velocity.inherit(velocity.x ?? 0, velocity.y ?? 0);
      }
      this.container.addChildAt(this.$object.container, 0);
      this.container.filters = filters;
   }
   draw (style: XStyle) {}
   render (camera: X2, style: XStyle) {
      if (this.$object.dirty) {
         this.$object.container.removeChildren();
         for (const object of this.objects) {
            this.$object.container.addChild(object.container);
         }
         this.$object.dirty = false;
      }
      if (this.gravity.extent.value !== 0) {
         Object.assign(
            this.velocity,
            this.velocity.endpoint(this.gravity.angle.value, this.gravity.extent.value * X.speed.value).value()
         );
      }
      if (this.friction.value !== 1) {
         const divisor = (this.friction.value - 1) * X.speed.value + 1;
         Object.assign(this.velocity, this.velocity.divide(divisor).value());
         this.spin.value /= divisor;
      }
      if (this.spin.value !== 0) {
         this.rotation.value += this.spin.value * X.speed.value;
         this.rotation.value > 360 && (this.rotation.value -= 360);
         this.rotation.value < 0 && (this.rotation.value += 360);
      }
      if (this.velocity.x !== 0 || this.velocity.y !== 0) {
         Object.assign(this.position, this.position.add(this.velocity.multiply(X.speed.value)));
      }
      this.fire('tick');
      const substyle = {
         blend: this.blend ?? style.blend,
         border: this.border ?? style.border,
         fill: this.fill ?? style.fill,
         font: this.font ?? style.font,
         stroke: this.stroke ?? style.stroke,
         tint: this.tint ?? style.tint
      };
      this.container.alpha = this.alpha.value;
      this.container.position.set(
         this.position.x + this.parallax.x * camera.x,
         this.position.y + this.parallax.y * camera.y
      );
      this.container.angle = this.rotation.value;
      this.container.scale.set(this.scale.x, this.scale.y);
      this.draw(substyle);
      for (const object of this.objects) {
         object.render(camera, substyle);
      }
      this.fire('render');
   }
}

/** An object with an anchor point. This is a base class for other anchored objects. Internal use only. */
export class XAnchoredObject extends XObject {
   /**
    * The anchor of the object. This value changes where the object is rendered relative to its `position` value. By default, this is `{ x: -1, y: -1 }`, which draws the object such that its top-left corner is at the object's `position`. When this value is `{ x: 0, y: 0 }`, the object is drawn centered on its `position`. Finally, a value of `{ x: 1, y: 1 }` will put its bottom-right corner at the object `position`.
    * @example
    * // A sprite centered on `position.x` and extending upward from `position.y`
    * const sprite = new XSprite({
    *    anchor: { x: 0, y: 1 }
    * });
    */
   anchor = new XVector();
   constructor (properties: XAnchoredObjectProperties = {}) {
      super(properties);
      const anchor = properties!.anchor ?? -1;
      if (typeof anchor === 'number') {
         this.anchor.inherit(anchor);
      } else {
         this.anchor.inherit(anchor.x ?? -1, anchor.y ?? -1);
      }
   }
   /** Set the object's `position` to the given value while retaining its displayed position. */
   cast (position: X2) {
      const size = this.compute();
      if (size.x !== 0 || size.y !== 0) {
         this.anchor.inherit(
            size
               .multiply(this.anchor)
               .subtract(this.position.multiply(2))
               .add(new XVector(position).multiply(2))
               .divide(size)
         );
      }
      this.position.inherit(position.x, position.y);
   }
   /** Returns the computed size of the object. */
   compute () {
      return new XVector();
   }
}

export class XSizedObject extends XAnchoredObject {
   /** The size of the object. */
   size = new XVector();
   constructor (properties: XSizedObjectProperties = {}) {
      super(properties);
      const size = properties!.size ?? 0;
      if (typeof size === 'number') {
         this.size.inherit(size);
      } else {
         this.size.inherit(size.x ?? 0, size.y ?? 0);
      }
   }
   compute () {
      return this.size;
   }
}

export class XHitbox extends XSizedObject {
   /** The state of the hitbox. Internal use only. */
   state = { dimensions: void 0, vertices: [] } as
      | { dimensions: void; vertices: [] }
      | { dimensions: string; vertices: [X2, X2, X2, X2] };
   region (): XRegion {
      const vertices = this.vertices();
      const { x: x1, y: y1 } = vertices[0];
      const { x: x2, y: y2 } = vertices[1];
      const { x: x3, y: y3 } = vertices[2];
      const { x: x4, y: y4 } = vertices[3];
      return [
         new XVector(Math.min(x1, x2, x3, x4), Math.min(y1, y2, y3, y4)),
         new XVector(Math.max(x1, x2, x3, x4), Math.max(y1, y2, y3, y4))
      ];
   }
   vertices () {
      if (this.state.dimensions === void 0) {
         throw "This object's vertices are unresolved!";
      } else {
         return this.state.vertices as [X2, X2, X2, X2];
      }
   }
}

/*
   export class XHitbox extends XObject {
      private $hitbox = {
         previous: [ 0, 0, 0, 0, 0, 0, 0, 0, 0 ] as [number, number, number, number, number, number, number, number, number],
         vertices: [
            { x: 0, y: 0 },
            { x: 0, y: 0 },
            { x: 0, y: 0 },
            { x: 0, y: 0 }
         ] as [X2, X2, X2, X2]
      };
      anchor = new XVector();
      size = new XVector();
      constructor (properties: XHitboxProperties = {}) {
         super(properties);
         (({
            anchor: { x: anchor_x = -1, y: anchor_y = -1 } = {},
            size: { x: size_x = 0, y: size_y = 0 } = {}
         }: XHitboxProperties = {}) => {
            this.anchor.inherit(anchor_x, anchor_y);
            this.size.inherit(size_x, size_y);
         })(properties);
      }
      detect (source: XHitbox, [ min1, max1 ] = source.region()) {
         const [ min2, max2 ] = this.region();
         if ((source.size.x === 0 || source.size.y === 0) && (this.size.x === 0 || this.size.y === 0)) {
            if (X.math.intersection(min1, max1, min2, max2)) {
               return true;
            }
         } else {
            if (min1.x < max2.x && min2.x < max1.x && min1.y < max2.y && min2.y < max1.y) {
               const vertices1 = source.vertices();
               const vertices2 = this.vertices();
               if (
                  (vertices1[0].x === vertices1[1].x || vertices1[0].y === vertices1[1].y) &&
                  (vertices2[0].x === vertices2[1].x || vertices2[0].y === vertices2[1].y)
               ) {
                  return true;
               } else {
                  for (const a1 of vertices1) {
                     let miss = true;
                     const a2 = new XVector(a1).add(new XVector(max2).subtract(min2).multiply(2)).value();
                     for (const [ b1, b2 ] of [
                        [ vertices2[0], vertices2[1] ],
                        [ vertices2[1], vertices2[2] ],
                        [ vertices2[2], vertices2[3] ],
                        [ vertices2[3], vertices2[0] ]
                     ]) {
                        if (X.math.intersection(a1, a2, b1, b2)) {
                           if ((miss = !miss)) {
                              break;
                           }
                        }
                     }
                     if (!miss) {
                        return true;
                     }
                  }
               }
            }
         }
         return false;
      }
      region (): XRegion {
         const { x: x1, y: y1 } = this.$hitbox.vertices[0];
         const { x: x2, y: y2 } = this.$hitbox.vertices[1];
         const { x: x3, y: y3 } = this.$hitbox.vertices[2];
         const { x: x4, y: y4 } = this.$hitbox.vertices[3];
         return [
            new XVector(Math.min(x1, x2, x3, x4), Math.min(y1, y2, y3, y4)),
            new XVector(Math.max(x1, x2, x3, x4), Math.max(y1, y2, y3, y4))
         ];
      }
      calculate ([ position, rotation, scale ]: XTransform) {
         let update = false;
         if (this.anchor.x !== this.$hitbox.previous[0]) {
            this.$hitbox.previous[0] = this.anchor.x;
            update = true;
         }
         if (this.anchor.y !== this.$hitbox.previous[1]) {
            this.$hitbox.previous[1] = this.anchor.y;
            update = true;
         }
         if (this.size.x !== this.$hitbox.previous[2]) {
            this.$hitbox.previous[2] = this.size.x;
            update = true;
         }
         if (this.size.y !== this.$hitbox.previous[3]) {
            this.$hitbox.previous[3] = this.size.y;
            update = true;
         }
         if (position.x !== this.$hitbox.previous[4]) {
            this.$hitbox.previous[4] = position.x;
            update = true;
         }
         if (position.y !== this.$hitbox.previous[5]) {
            this.$hitbox.previous[5] = position.y;
            update = true;
         }
         if (rotation !== this.$hitbox.previous[6]) {
            this.$hitbox.previous[6] = rotation;
            update = true;
         }
         if (scale.x !== this.$hitbox.previous[7]) {
            this.$hitbox.previous[7] = scale.x;
            update = true;
         }
         if (scale.y !== this.$hitbox.previous[8]) {
            this.$hitbox.previous[8] = scale.y;
            update = true;
         }
         if (update) {
            rotation += 180;
            const size = this.size.multiply(scale);
            const half = size.divide(2);
            const base = position.subtract(half.add(half.multiply(this.anchor)));
            this.$hitbox.vertices[0] = position.shift(rotation, 0, base).round(1e6);
            this.$hitbox.vertices[1] = position.shift(rotation, 0, base.add(size.x, 0)).round(1e6);
            this.$hitbox.vertices[2] = position.shift(rotation, 0, base.add(size.x, size.y)).round(1e6);
            this.$hitbox.vertices[3] = position.shift(rotation, 0, base.add(0, size.y)).round(1e6);
         }
      }
      vertices () {
         return this.$hitbox.vertices;
      }
   }
   */

export class XEntity extends XHitbox {
   face: XCardinal = 'down';
   sprites: XKeyed<XSprite, XCardinal>;
   step: number;
   get sprite () {
      return this.sprites[this.face];
   }
   constructor (properties: XEntityProperties = {}) {
      super(properties);
      (({
         sprites: { down = new XSprite(), left = new XSprite(), right = new XSprite(), up = new XSprite() } = {},
         step = 1
      }: XEntityProperties = {}) => {
         this.sprites = { down, left, right, up };
         this.step = step;
      })(properties);
   }
   move<A extends string> (
      offset: X2,
      renderer?: XRenderer<A>,
      key?: A,
      keys: A[] = [],
      filter?: XProvider<boolean, [XHitbox]>
   ) {
      const source = this.position.value();
      const hitboxes = filter && renderer ? keys.flatMap(key => renderer.calculate(key, filter)) : [];
      for (const axis of [ 'x', 'y' ] as ['x', 'y']) {
         const distance = offset[axis];
         if (distance !== 0) {
            this.position[axis] += distance;
            const hits = renderer ? renderer.detect(key, this, ...hitboxes) : [];
            if (hits.length > 0) {
               const single = (distance / Math.abs(distance)) * this.step;
               while (this.position[axis] !== source[axis] && renderer!.detect(key, this, ...hits).length > 0) {
                  this.position[axis] -= single;
               }
            }
         }
      }
      if (this.position.x === source.x && this.position.y === source.y) {
         if (offset.y < 0) {
            this.face = 'up';
         } else if (offset.y > 0) {
            this.face = 'down';
         } else if (offset.x < 0) {
            this.face = 'left';
         } else if (offset.x > 0) {
            this.face = 'right';
         }
         for (const sprite of Object.values(this.sprites)) {
            sprite.reset();
         }
         return false;
      } else {
         if (this.position.y < source.y) {
            this.face = 'up';
         } else if (this.position.y > source.y) {
            this.face = 'down';
         } else if (this.position.x < source.x) {
            this.face = 'left';
         } else if (this.position.x > source.x) {
            this.face = 'right';
         }
         this.sprite.enable();
         return true;
      }
   }
   render (camera: X2, style: XStyle) {
      this.objects[0] = this.sprites[this.face];
      super.render(camera, style);
   }
   async walk<A extends string> (renderer: XRenderer<A>, key: A, speed: number, ...points: Partial<X2>[]) {
      await renderer.on('tick');
      for (const sprite of Object.values(this.sprites)) {
         sprite.duration = Math.round(15 / speed);
      }
      for (const { x = this.position.x, y = this.position.y } of points) {
         const dirX = x - this.position.x < 0 ? -1 : 1;
         const dirY = y - this.position.y < 0 ? -1 : 1;
         await X.chain<void, Promise<void>>(void 0, async (z, next) => {
            const diffS = speed * X.speed.value;
            const diffX = Math.abs(x - this.position.x);
            const diffY = Math.abs(y - this.position.y);
            this.move(
               {
                  x: (diffX === 0 ? 0 : diffX < diffS ? diffX : diffS) * dirX,
                  y: (diffY === 0 ? 0 : diffY < diffS ? diffY : diffS) * dirY
               },
               renderer,
               key
            );
            if (diffX > 0 || diffY > 0) {
               await renderer.on('tick');
               await next();
            }
         });
      }
   }
}

export class XCharacter extends XEntity {
   key: string;
   preset: XCharacterPreset;
   talk = false;
   get sprite () {
      return (this.talk ? this.preset.talk : this.preset.walk)[this.face];
   }
   constructor (properties: XCharacterProperties) {
      super(properties);
      (({ key, preset }: XCharacterProperties) => {
         this.key = key;
         this.preset = preset;
      })(properties);
   }
   render (camera: X2, style: XStyle) {
      this.sprites = this.talk ? this.preset.talk : this.preset.walk;
      super.render(camera, style);
   }
}

export class XPlayer extends XEntity {
   private $player = { extension: new XHitbox(), face: null as XCardinal | null };
   extent = new XVector();
   override = false;
   constructor (properties: XEntityProperties & XProperties<XPlayer, 'extent'> = {}) {
      super(properties);
      const extent = properties!.extent ?? -1;
      if (typeof extent === 'number') {
         this.extent.inherit(extent);
      } else {
         this.extent.inherit(extent.x ?? -1, extent.y ?? -1);
      }
   }
   render (camera: X2, style: XStyle) {
      if (this.face !== this.$player.face) {
         this.$player.face = this.face;
         this.$player.extension.anchor.inherit(
            this.face === 'left' ? 1 : this.face === 'right' ? -1 : 0,
            this.face === 'up' ? 1 : this.face === 'down' ? -1 : 0
         );
         this.$player.extension.size.inherit(
            this.face === 'left' || this.face === 'right' ? this.extent.y : this.extent.x,
            this.face === 'down' || this.face === 'up' ? this.extent.y : this.extent.x
         );
      }
      this.objects[1] = this.$player.extension;
      super.render(camera, style);
   }
   async walk<A extends string> (renderer: XRenderer<A>, key: A, speed: number, ...targets: Partial<X2>[]) {
      this.override = true;
      await super.walk(renderer, key, speed, ...targets);
      this.override = false;
   }
}

export class XRandom {
   /** The current base value of the generator. */
   base: number;
   constructor (
      seed: string,
      base = (() => {
         let h = 1779033703 ^ seed.length;
         for (let i = 0; i < seed.length; i++) {
            (h = Math.imul(h ^ seed.charCodeAt(i), 3432918353)), (h = (h << 13) | (h >>> 19));
         }
         let [ a, b, c, d ] = X.populate(4, () => {
            h = Math.imul(h ^ (h >>> 16), 2246822507);
            h = Math.imul(h ^ (h >>> 13), 3266489909);
            return (h ^= h >>> 16) >>> 0;
         });
         a >>>= 0;
         b >>>= 0;
         c >>>= 0;
         d >>>= 0;
         let t = (a + b) | 0;
         a = b ^ (b >>> 9);
         b = (c + (c << 3)) | 0;
         c = (c << 21) | (c >>> 11);
         d = (d + 1) | 0;
         t = (t + d) | 0;
         c = (c + t) | 0;
         return (t >>> 0) / 4294967296;
      })()
   ) {
      this.base = base;
   }
   /** Advances the generator forward by a step, returning the next random number. If a `threshold` is specified, the difference between the new and old random number must be greater than it. */
   next (
      /* The difference threshold, with a valid range of `[0, 0.5]` */
      threshold = 0
   ) {
      const base = this.base;
      let step = this.step();
      const trueThreshold = Math.min(Math.max(threshold, 0), 0.5);
      while (Math.abs(this.base - base) < trueThreshold) {
         step = this.step();
      }
      return step;
   }
   /** Returns the next random number in the generator. */
   step () {
      let hash = (this.base += 0x6d2b79f5);
      hash = Math.imul(hash ^ (hash >>> 15), hash | 1);
      hash ^= hash + Math.imul(hash ^ (hash >>> 7), hash | 61);
      return ((hash ^ (hash >>> 14)) >>> 0) / 4294967296;
   }
}

export class XRectangle extends XSizedObject {
   $rectangle = { graphics: new PIXI.Graphics(), visible: false };
   draw (style: XStyle) {
      if (this.container.alpha > 0) {
         if (!this.$rectangle.visible) {
            this.$rectangle.visible = true;
            this.container.addChildAt(this.$rectangle.graphics, 0);
         }
         const graphics = this.$rectangle.graphics.clear();
         const fill = style.fill !== 'transparent';
         const stroke = style.stroke !== 'transparent';
         if (fill || stroke) {
            const half = this.size.divide(2);
            const base = half.add(half.multiply(this.anchor));
            graphics.pivot.set(this.size.x * ((this.anchor.x + 1) / 2), this.size.y * ((this.anchor.y + 1) / 2));
            graphics.position.set(base.x - graphics.pivot.x, base.y - graphics.pivot.y);
            graphics.blendMode = X.blend(style.blend);
            if (stroke) {
               const strokeColor = X.cache.color.of(style.stroke);
               if (strokeColor[3] > 0) {
                  graphics.lineStyle({
                     alpha: strokeColor[3] * this.alpha.value,
                     color: X.color2hex(strokeColor),
                     width: style.border
                  });
               }
            }
            if (fill) {
               const fillColor = X.cache.color.of(style.fill);
               if (fillColor[3] > 0) {
                  graphics.beginFill(X.color2hex(fillColor), fillColor[3] * this.alpha.value);
               } else {
                  graphics.beginFill(0, 0);
               }
            }
            graphics.drawRect(0, 0, this.size.x, this.size.y).endFill();
            graphics.tint = style.tint;
         }
      } else if (this.$rectangle.visible) {
         this.$rectangle.visible = false;
         this.container.removeChildAt(0);
      }
   }
}

/** A renderer. */
export class XRenderer<A extends string = string> extends XHost<{ tick: []; render: [A] }> {
   alpha = new XNumber();
   application: PIXI.Application;
   camera = new XVector();
   container: HTMLElement;
   layers: XKeyed<XRendererLayer, A>;
   private quality = new XNumber();
   region: XRegion;
   rotation = new XNumber();
   shake = new XNumber();
   private size = new XVector();
   state = {
      camera: { x: NaN, y: NaN },
      active: false,
      rendererX: NaN,
      rendererY: NaN,
      scale: 1,
      windowX: NaN,
      windowY: NaN,
      quality: NaN,
      zoom: NaN
   };
   zoom = new XNumber();
   constructor ({
      alpha = 1,
      camera = 0,
      container = document.body,
      layers = {} as XKeyed<XRendererLayer, A>,
      region: [
         { x: min_x = -Infinity, y: min_y = -Infinity } = {},
         { x: max_x = Infinity, y: max_y = Infinity } = {}
      ] = [],
      rotation = 0,
      shake = 0,
      size = 1,
      quality = 1,
      zoom = 1
   }: XRendererProperties<A> = {}) {
      super();
      const element = typeof container === 'string' ? (document.querySelector(container) as HTMLElement) : container;
      this.alpha.value = alpha;
      if (typeof camera === 'number') {
         this.camera.inherit(camera);
      } else {
         this.camera.inherit(camera.x ?? 0, camera.y ?? 0);
      }
      this.container = element;
      this.layers = layers;
      this.quality.value = quality;
      this.region = [
         { x: min_x, y: min_y },
         { x: max_x, y: max_y }
      ];
      this.rotation.value = rotation;
      this.shake.value = shake;
      if (typeof size === 'number') {
         this.size.inherit(size);
      } else {
         this.size.inherit(size.x ?? 0, size.y ?? 0);
      }
      this.zoom.value = zoom;
      this.application = new PIXI.Application({
         antialias: false,
         autoStart: false,
         backgroundAlpha: 0,
         height: this.size.y * this.quality.value,
         width: this.size.x * this.quality.value
      });
      (this.application.renderer.view as HTMLCanvasElement).style.imageRendering = 'pixelated';
      this.application.ticker.add(() => this.render());
      this.application.ticker.minFPS = 30;
      this.application.ticker.maxFPS = 30;
      this.container.appendChild(this.application.renderer.view as HTMLCanvasElement);
      for (const key in this.layers) {
         this.application.stage.addChild(this.layers[key].container);
      }
   }
   attach (key: A, ...objects: XObject[]) {
      if (key in this.layers) {
         const layer = this.layers[key];
         layer.modifiers.includes('ambient') && this.refresh();
         for (const object of objects) {
            layer.objects.includes(object) || layer.objects.push(object);
         }
         layer.container.removeChildren();
         for (const object of layer.objects.sort(
            (object1, object2) => object1.priority.value - object2.priority.value
         )) {
            layer.container.addChild(object.container);
            X.chain(object, (object, next) => {
               const subobjects = object.objects.slice();
               subobjects.sort((object1, object2) => object1.priority.value - object2.priority.value);
               for (const subobject of (object.objects = subobjects)) {
                  next(subobject);
               }
            });
         }
      }
   }
   /*
      calculate(source: A | XObject[], filter: XProvider<boolean, [XHitbox]>): XHitbox[];
      calculate(
         source: A | XObject[],
         filter: XProvider<boolean, [XHitbox]>,
         transform: XTransform,
         camera: X2,
         sublist: XHitbox[]
      ): void;
      calculate (
         source: A | XObject[],
         filter: XProvider<boolean, [XHitbox]>,
         transform: XTransform = [ new XVector(), 0, new XVector(0.5) ],
         camera: X2 = this.camera.clamp(...this.region),
         sublist?: XHitbox[]
      ) {
         const list = sublist ?? ([] as XHitbox[]);
         const objects = typeof source === 'string' ? this.layers[source].objects : source;
         for (const object of objects) {
            const subtransform: XTransform = [
               transform[0].add(object.position).add(object.parallax.multiply(camera)),
               transform[1] + object.rotation.value,
               transform[2].multiply(object.scale)
            ];
            if (object instanceof XHitbox && X.provide(filter, object)) {
               list.push(object);
               object.calculate(subtransform);
            }
            this.calculate(object.objects, filter, subtransform, camera, list);
         }
         if (!sublist) {
            return list;
         }
      }
      */
   calculate (key: A, filter: XProvider<boolean, [XHitbox]> = true) {
      const list: XHitbox[] = [];
      const camera = this.camera.clamp(...this.region);
      X.chain(
         [ [ new XVector(), 0, new XVector(1) ], this.layers[key].objects ] as [XTransform, XObject[]],
         ([ transform, objects ], next) => {
            for (const object of objects) {
               const position = transform[0].add(object.position).add(object.parallax.multiply(camera));
               const rotation = transform[1] + object.rotation.value;
               const scale = transform[2].multiply(object.scale);
               if (object instanceof XHitbox && X.provide(filter, object)) {
                  list.push(object);
                  const size = object.size.multiply(scale);
                  const half = size.divide(2);
                  const base = position.subtract(half.add(half.multiply(object.anchor)));
                  const dimensions = `${base.x}:${base.y}:${position.x}:${position.y}:${rotation}:${size.x}:${size.y}`;
                  if (dimensions !== object.state.dimensions) {
                     const offset = rotation + 180;
                     const corner2 = base.endpoint(0, size.x);
                     const corner3 = corner2.endpoint(90, size.y);
                     const corner4 = corner3.endpoint(180, size.x);
                     object.state.vertices[0] = position
                        .endpoint(position.angleFrom(base) + offset, position.extentOf(base))
                        .round(1e6);
                     object.state.vertices[1] = position
                        .endpoint(position.angleFrom(corner2) + offset, position.extentOf(corner2))
                        .round(1e6);
                     object.state.vertices[2] = position
                        .endpoint(position.angleFrom(corner3) + offset, position.extentOf(corner3))
                        .round(1e6);
                     object.state.vertices[3] = position
                        .endpoint(position.angleFrom(corner4) + offset, position.extentOf(corner4))
                        .round(1e6);
                     object.state.dimensions = dimensions;
                  }
               }
               next([ [ position, rotation, scale ], object.objects ]);
            }
         }
      );
      return list;
   }
   clear (key: A) {
      if (key in this.layers) {
         const layer = this.layers[key];
         layer.modifiers.includes('ambient') && this.refresh();
         for (const object of layer.objects.splice(0, layer.objects.length)) {
            layer.container.removeChild(object.container);
         }
      }
   }
   detach (key: A, ...objects: XObject[]) {
      if (key in this.layers) {
         const layer = this.layers[key];
         layer.modifiers.includes('ambient') && this.refresh();
         for (const object of objects) {
            const index = layer.objects.indexOf(object);
            if (index > -1) {
               layer.objects.splice(index, 1);
               layer.container.removeChild(object.container);
            }
         }
      }
   }
   /*
      detect (key: A | void, source: XHitbox, ...hitboxes: XHitbox[]) {
         key && this.calculate(key, hitbox => hitbox === source);
         const region = source.region();
         return hitboxes.filter(hitbox => hitbox.detect(source, region));
      }
      */
   detect (key: A | void, source: XHitbox, ...hitboxes: XHitbox[]) {
      key && this.calculate(key, hitbox => hitbox === source);
      const hits: XHitbox[] = [];
      const [ min1, max1 ] = source.region();
      for (const hitbox of hitboxes) {
         if (hitbox.state.dimensions === void 0) {
            continue;
         } else if ((source.size.x === 0 || source.size.y === 0) && (hitbox.size.x === 0 || hitbox.size.y === 0)) {
            const [ min2, max2 ] = hitbox.region();
            if (X.math.intersection(min1, max1, min2, max2)) {
               hits.push(hitbox);
            }
         } else {
            const [ min2, max2 ] = hitbox.region();
            if (min1.x < max2.x && min2.x < max1.x && min1.y < max2.y && min2.y < max1.y) {
               const vertices1 = source.vertices();
               const vertices2 = hitbox.vertices();
               if (
                  (vertices1[0].x === vertices1[1].x || vertices1[0].y === vertices1[1].y) &&
                  (vertices2[0].x === vertices2[1].x || vertices2[0].y === vertices2[1].y)
               ) {
                  hits.push(hitbox);
               } else {
                  for (const a1 of vertices1) {
                     let miss = true;
                     const a2 = new XVector(a1).add(new XVector(max2).subtract(min2).multiply(2)).value();
                     for (const [ b1, b2 ] of [
                        [ vertices2[0], vertices2[1] ],
                        [ vertices2[1], vertices2[2] ],
                        [ vertices2[2], vertices2[3] ],
                        [ vertices2[3], vertices2[0] ]
                     ]) {
                        if (X.math.intersection(a1, a2, b1, b2)) {
                           if ((miss = !miss)) {
                              break;
                           }
                        }
                     }
                     if (!miss) {
                        hits.push(hitbox);
                        break;
                     }
                  }
               }
            }
         }
      }
      return hits;
   }
   async pause (time: number) {
      let elapsed = 0;
      const duration = Math.round((time * 0.03) / X.speed.value);
      while (elapsed++ < duration) {
         await this.on('tick');
      }
   }
   refresh () {
      this.state.camera = { x: NaN, y: NaN };
   }
   restrict (position: X2) {
      return this.size
         .divide(2)
         .subtract(this.camera.clamp(...this.region))
         .add(position);
   }
   render () {
      this.fire('tick');
      let update = false;
      let resize = false;
      const camera = this.camera.clamp(...this.region);
      this.container.style.opacity = Math.min(Math.max(this.alpha.value, 0), 1).toString();
      if (this.zoom.value !== this.state.zoom) {
         update = true;
         this.state.zoom = this.zoom.value;
      }
      if (camera.x !== this.state.camera.x || camera.y !== this.state.camera.y) {
         update = true;
         this.state.camera = camera.value();
      }
      if (this.quality.value !== this.state.quality) {
         update = true;
         resize = true;
         this.state.quality = this.quality.value;
      }
      const innerWidth = this.container.clientWidth;
      const innerHeight = this.container.clientHeight;
      if (
         innerWidth !== this.state.windowX ||
         innerHeight !== this.state.windowY ||
         this.size.x !== this.state.rendererX ||
         this.size.y !== this.state.rendererY
      ) {
         update = true;
         resize = true;
         this.state.windowX = innerWidth;
         this.state.windowY = innerHeight;
         this.state.rendererX = this.size.x;
         this.state.rendererY = this.size.y;
         const ratio = this.size.x / this.size.y;
         if (this.state.windowX / this.state.windowY > ratio) {
            this.state.scale = innerHeight / this.size.y;
         } else {
            this.state.scale = innerWidth / this.size.x;
         }
      }
      if (resize) {
         const width = this.size.x * this.quality.value;
         const height = this.size.y * this.quality.value;
         const scale = this.state.scale / this.quality.value;
         {
            const canvas = this.application.renderer.view as HTMLCanvasElement;
            canvas.width = width;
            canvas.height = height;
            canvas.style.transform = `scale(${scale})`;
         }
      }
      let shakeX = 0;
      let shakeY = 0;
      if (this.shake.value !== 0) {
         update = true;
         shakeX = Math.abs(this.shake.value) * (Math.random() - 0.5);
         shakeY = Math.abs(this.shake.value) * (Math.random() - 0.5);
      }
      const center = this.size.divide(2);
      for (const key in this.layers) {
         const { active, container, modifiers, objects } = this.layers[key];
         if (active && (update || !modifiers.includes('ambient'))) {
            const secure = modifiers.includes('secure');
            const statik = modifiers.includes('static');
            if (modifiers.includes('vertical')) {
               for (const object of objects) {
                  object.container.zIndex = object.priority.value;
               }
               container.children.sort(
                  (child1, child2) => (child1.zIndex || child1.position.y) - (child2.zIndex || child2.position.y)
               );
            }
            const zoom = statik ? 1 : this.zoom.value;
            const subcamera = camera.add(this.size.subtract(this.size.divide(zoom)).divide(2));
            const factor = zoom * this.quality.value;
            container.position.set(
               (statik ? 0 : (center.x - subcamera.x) * factor) + (secure ? 0 : shakeX),
               (statik ? 0 : (center.y - subcamera.y) * factor) + (secure ? 0 : shakeY)
            );
            container.angle = statik ? 0 : this.rotation.value;
            container.scale.set(factor);
            container.filterArea = new PIXI.Rectangle(
               0,
               0,
               this.size.x * this.quality.value,
               this.size.y * this.quality.value
            );
            for (const object of objects) {
               object.render(camera, X.style);
            }
            this.fire('render', key);
         }
      }
   }
   start () {
      this.application.start();
      return () => this.application.stop();
   }
}

/** A rendering layer. */
export class XRendererLayer implements XObjectParent {
   /** Controls whether or not this layer should be rendered. Defaults to `true`. */
   active: boolean;
   /** The layer's PIXI container. */
   container = new PIXI.Container();
   /** The layer's modifiers. Refer to the `XRendererLayerModifier` type for a breakdown of how each modifier value behaves. */
   modifiers: XRendererLayerModifier[];
   /** The layer's stored objects. This is readonly to prevent untracked object list manipulation. */
   readonly objects: XObject[] = [];
   constructor ({ active = true, filters = [], modifiers = [] }: XRendererLayerProperties = {}) {
      this.active = active;
      this.modifiers = modifiers;
      this.container.filters = filters;
   }
}

export class XSprite extends XAnchoredObject {
   private $sprite = {
      graphics: new PIXI.Graphics(),
      image: null as XImage | null,
      sprite: new PIXI.Sprite(),
      visible: false
   };
   /** The active state of the sprite. While true, the sprite will iterate over its frames. */
   active = false;
   /** The sprite.s crop. Refer to the `XCrop` type for more information. */
   crop: XCrop;
   /** The sprite's duration. A value of `1` is equal to 1/30th of a second. */
   duration: number;
   /** The sprite's frames. The sprite's `index` value is used too choose which of these is drawn. If there are no frames available, or the current frame is `null`, then nothing is drawn. */
   frames: XFrames;
   /** The sprite's increment. When a frame completes, the index will be shifted by this value. */
   increment: number;
   /** The sprite's index. The image at this index of the `frames` array is what gets rendered. */
   index = 0;
   /** The sprite's step offset, or in other words, the default `step` value of the sprite. When the sprite is initialized or reset, the sprite's `step` is set to this value. */
   offset: number;
   /** The sprite's step. This value is incremented once per frame, only while the sprite is active. When it meets or exceeds the sprite's `duration`, the sprite moves to the next frame. */
   step = 0;
   /** The sprite's internal graphics object, used to apply `crop` when rendered. */
   get graphics () {
      return this.$sprite.graphics;
   }
   get sprite () {
      return this.$sprite.sprite;
   }
   constructor (properties: XSpriteProperties = {}) {
      super(properties);
      (({
         active = false,
         crop: { bottom = 0, left = 0, right = 0, top = 0 } = {},
         duration = 1,
         frames = [],
         increment = 1,
         index = 0,
         offset = 0
      }: XSpriteProperties = {}) => {
         this.active = active;
         this.crop = { bottom, left, right, top };
         this.duration = duration;
         this.frames = frames;
         this.increment = increment;
         this.index = index;
         this.offset = offset;
      })(properties);
   }
   compute () {
      if (this.container.alpha > 0) {
         const image = this.frames[this.index];
         if (image) {
            if (image.state.value) {
               const crop = X.crop(image.value, this.crop);
               return new XVector(crop.w, crop.h);
            } else {
               X.status(`Error: Cannot invoke the unloaded asset of "${image.source}"`, {
                  backgroundColor: '#f00',
                  color: '#fff',
                  fontSize: '12px'
               });
            }
         }
      }
      return new XVector();
   }
   /** Sets the sprite as inactive. */
   disable () {
      this.active = false;
      return this;
   }
   draw (style: XStyle) {
      if (this.container.alpha > 0) {
         if (!this.$sprite.visible) {
            this.$sprite.visible = true;
            this.container.addChildAt(this.$sprite.graphics, 0);
            this.container.addChildAt(this.$sprite.sprite, 1);
         }
         const sprite = this.$sprite.sprite;
         const graphics = this.$sprite.graphics.clear();
         const prev = this.$sprite.image;
         const next = this.frames[this.index];
         if (prev !== next) {
            if (prev && X.cache.sprites.has(prev)) {
               const sprites = X.cache.sprites.get(prev)!;
               sprites.delete(sprite);
               if (sprites.size === 0) {
                  X.cache.sprites.delete(prev);
               }
            }
            if (next) {
               X.cache.sprites.of(next).add(sprite);
               if (next.state.value) {
                  sprite.texture = next.value;
               } else {
                  X.status(`Error: Cannot invoke the unloaded asset of "${next.source}"`, {
                     backgroundColor: '#f00',
                     color: '#fff',
                     fontSize: '12px'
                  });
               }
            } else {
               sprite.texture = X.texture;
            }
            this.$sprite.image = next;
         }
         if (sprite.texture.baseTexture) {
            const a = (this.anchor.x + 1) / 2;
            const b = (this.anchor.y + 1) / 2;
            const { x, y, w, h } = X.crop(sprite.texture, this.crop);
            sprite.anchor.set((x + w * a) / sprite.texture.width, (y + h * b) / sprite.texture.height);
            sprite.blendMode = X.blend(style.blend);
            graphics.pivot.set(w * a, h * b);
            graphics.beginFill(0xff0000, 1).drawRect(0, 0, w, h).endFill();
            const mask = new PIXI.MaskData(graphics);
            mask.type = sprite.angle % 90 === 0 ? PIXI.MASK_TYPES.SCISSOR : PIXI.MASK_TYPES.STENCIL;
            sprite.mask = mask;
            sprite.tint = style.tint;
         } else {
            sprite.texture = X.texture;
         }
      } else if (this.$sprite.visible) {
         this.$sprite.visible = false;
         this.container.removeChildAt(1);
         this.container.removeChildAt(0);
      }
   }
   /** Sets the sprite as acctive. */
   enable (duration?: number) {
      this.active = true;
      duration && (this.duration = duration);
      return this;
   }
   /** Extracts a portion of this image. */
   read (): XColor[][] | null {
      const image = this.frames[this.index];
      if (image) {
         if (image.state.value) {
            const sprite = PIXI.Sprite.from(image.value);
            const pixels = X.chalkboard.pixels(sprite);
            sprite.destroy();
            const { x, y, w, h } = X.crop(image.value, this.crop);
            const width = image.value.width;
            const index = { x: 0, y: 0 };
            const colors: XColor[][] = [];
            while (index.x < w) {
               const subcolors = (colors[index.x] = [] as XColor[]);
               while (index.y < h) {
                  const value = Math.floor(x + index.x + (y + index.y++) * width) * 4;
                  subcolors.push([ pixels[value], pixels[value + 1], pixels[value + 2], pixels[value + 3] ] as XColor);
               }
               index.y = 0;
               index.x++;
            }
            return colors;
         } else {
            X.status(`Error: Cannot invoke the unloaded asset of "${image.source}"`, {
               backgroundColor: '#f00',
               color: '#fff',
               fontSize: '12px'
            });
         }
      }
      return null;
   }
   render (camera: X2, style: XStyle) {
      super.render(camera, style);
      if (this.active) {
         if (Math.round(this.duration / X.speed.value) <= ++this.step) {
            this.step = 0;
            this.index += this.increment;
            if (this.frames.length <= this.index) {
               this.index = 0;
            } else if (this.index < 0) {
               this.index = this.frames.length - 1;
            }
         }
      }
   }
   /** Sets the sprite as inactive, and resets its index and step values. */
   reset (index = 0) {
      this.active = false;
      this.index = index;
      this.step = this.offset;
      return this;
   }
}

export class XAnimation extends XSprite {
   private $animation = { index: NaN };
   /** Controls whether or not the `steps` property of an animation will be automatically updated based on the duration of the current frame according to the animation data. In other words, the animation's timing will match what is exported from Aseprite. Defaults to `true`. */
   extrapolate: boolean;
   /** Controls an animation's resources. Refer to the `XAnimationResources` type for an explanation of how to generate and import these kinds of assets. */
   resources: XAnimationResources | null;
   /** Controls how an animation is cropped. This property replaces the `crop` property on sprites. Refer to the `XCrop` type for a detailed breakdown on cropping behavior. */
   subcrop: XCrop;
   constructor (properties: XAnimationProperties = {}) {
      super(properties);
      (({
         extrapolate = true,
         resources,
         subcrop: { left = 0, right = 0, bottom = 0, top = 0 } = {}
      }: XAnimationProperties = {}) => {
         this.extrapolate = extrapolate;
         this.resources = resources ?? null;
         this.subcrop = { left, right, bottom, top };
      })(properties);
   }
   refresh (lazy = false) {
      if (this.resources) {
         if (this.resources.state.value) {
            const frames = this.resources.value[0].value.frames;
            if (this.index < frames.length) {
               const value = frames[this.index];
               if (value) {
                  const { duration, frame } = frames[this.index];
                  const { x, y, w, h } = X.crop({ height: frame.h, width: frame.w }, this.subcrop);
                  this.crop = {
                     left: frame.x + x,
                     top: frame.y + y,
                     right: -(frame.x + x + w),
                     bottom: -(frame.y + y + h)
                  };
                  if (!lazy || this.index !== this.$animation.index) {
                     this.$animation.index = this.index;
                     const content = this.resources.value[1];
                     this.frames = X.populate(frames.length, () => content);
                     if (this.extrapolate) {
                        this.duration = Math.round(duration / (1 / 0.03));
                     }
                  }
               }
            }
         } else if (this.container.alpha > 0) {
            X.status(`Error: Cannot invoke the unloaded asset of "${this.resources.source}"`, {
               backgroundColor: '#f00',
               color: '#fff',
               fontSize: '12px'
            });
         }
      }
      return this;
   }
   render (camera: X2, style: XStyle) {
      this.refresh(true);
      super.render(camera, style);
   }
   update (resources: XAnimationResources | null) {
      this.resources = resources;
      if (this.active) {
         return this.reset().refresh().enable();
      } else {
         return this.reset().refresh();
      }
   }
   static resources (dataSource: string, imageSource: string, shader?: XShader): XAnimationResources {
      return X.inventory(X.dataAsset(dataSource), X.imageAsset(imageSource, { shader }));
   }
}

export class XVector implements X2 {
   x: number;
   y: number;
   get angle () {
      return this.angleFrom(0);
   }
   set angle (value) {
      this.inherit(new XVector().endpoint(value, this.extent));
   }
   get extent () {
      return this.extentOf(0);
   }
   set extent (value) {
      this.inherit(new XVector().endpoint(this.angle, value));
   }
   constructor();
   constructor(a: number | X2);
   constructor(x: number, y: number);
   constructor (a: number | X2 = 0, b = a as number) {
      if (typeof a === 'number') {
         this.x = a;
         this.y = b;
      } else {
         this.x = a.x ?? 0;
         this.y = a.y ?? 0;
      }
   }
   add(a: number | X2): XVector;
   add(x: number, y: number): XVector;
   add (a: number | X2, b = a as number) {
      if (typeof a === 'number') {
         return new XVector(this.x + a, this.y + b);
      } else {
         return this.add(a.x, a.y);
      }
   }
   angleFrom(a: number | X2): number;
   angleFrom(x: number, y: number): number;
   angleFrom (a: number | X2, b = a as number) {
      if (typeof a === 'number') {
         if (this.y === b) {
            if (this.x < a) {
               return 180;
            } else {
               return 0;
            }
         } else if (this.x === a) {
            if (this.y < b) {
               return 270;
            } else {
               return 90;
            }
         } else {
            return (Math.atan2(this.y - b, this.x - a) * (180 / Math.PI)) % 360;
         }
      } else {
         return this.angleFrom(a.x, a.y);
      }
   }
   angleTo(a: number | X2): number;
   angleTo(x: number, y: number): number;
   angleTo (a: number | X2, b = a as number) {
      if (typeof a === 'number') {
         return (this.angleFrom(a, b) + 180) % 360;
      } else {
         return this.angleTo(a.x, a.y);
      }
   }
   ceil (base?: number): XVector {
      return base ? this.multiply(base).ceil().divide(base) : new XVector(Math.ceil(this.x), Math.ceil(this.y));
   }
   clamp (min: X2, max: X2) {
      return new XVector(Math.min(Math.max(this.x, min.x), max.x), Math.min(Math.max(this.y, min.y), max.y));
   }
   clone () {
      return new XVector(this);
   }
   divide(a: number | X2): XVector;
   divide(x: number, y: number): XVector;
   divide (a: number | X2, b = a as number) {
      if (typeof a === 'number') {
         return new XVector(this.x / a, this.y / b);
      } else {
         return this.divide(a.x, a.y);
      }
   }
   endpoint (angle: number, extent: number) {
      if (extent === 0) {
         return this.clone();
      } else {
         switch (angle % 360) {
            case 0:
               return this.add(extent, 0);
            case 90:
            case -270:
               return this.add(0, extent);
            case 180:
            case -180:
               return this.subtract(extent, 0);
            case 270:
            case -90:
               return this.subtract(0, extent);
            default:
               const rads = Math.PI - (((angle + 90) % 360) * Math.PI) / 180;
               return new XVector(this.x + extent * Math.sin(rads), this.y + extent * Math.cos(rads));
         }
      }
   }
   extentOf(a: number | X2): number;
   extentOf(x: number, y: number): number;
   extentOf (a: number | X2, b = a as number) {
      if (typeof a === 'number') {
         if (this.y === b) {
            return Math.abs(a - this.x);
         } else if (this.x === a) {
            return Math.abs(b - this.y);
         } else {
            return Math.sqrt((a - this.x) ** 2 + (b - this.y) ** 2);
         }
      } else {
         return this.extentOf(a.x, a.y);
      }
   }
   floor (base?: number): XVector {
      return base ? this.multiply(base).floor().divide(base) : new XVector(Math.floor(this.x), Math.floor(this.y));
   }
   inherit(a: number | X2): XVector;
   inherit(x: number, y: number): XVector;
   inherit (a: number | X2, b = a as number) {
      if (typeof a === 'number') {
         this.x = a;
         this.y = b;
         return this;
      } else {
         return this.inherit(a.x, a.y);
      }
   }
   interpolate (duration: number, ...points: Partial<X2>[]) {
      const x = this.x;
      const y = this.y;
      const base = X.timer.value;
      let active = true;
      X.cache.tasks.get(this)?.();
      X.cache.tasks.set(this, () => (active = false));
      return X.when(() => {
         if (active) {
            const elapsed = X.timer.value - base;
            if (elapsed < duration) {
               this.x = X.math.linear(elapsed / duration, x, ...points.map(point => point.x ?? x));
               this.y = X.math.linear(elapsed / duration, y, ...points.map(point => point.y ?? y));
               return false;
            } else {
               X.cache.tasks.delete(this);
               this.x = points.length > 0 ? points[points.length - 1].x ?? x : x;
               this.y = points.length > 0 ? points[points.length - 1].y ?? y : y;
            }
         }
         return true;
      });
   }
   modulate (duration: number, ...points: Partial<X2>[]) {
      const x = this.x;
      const y = this.y;
      const base = X.timer.value;
      let active = true;
      X.cache.tasks.get(this)?.();
      X.cache.tasks.set(this, () => (active = false));
      return X.when(() => {
         if (active) {
            const elapsed = X.timer.value - base;
            if (elapsed < duration) {
               this.x = X.math.bezier(elapsed / duration, x, ...points.map(point => point.x ?? x));
               this.y = X.math.bezier(elapsed / duration, y, ...points.map(point => point.y ?? y));
               return false;
            } else {
               X.cache.tasks.delete(this);
               this.x = points.length > 0 ? points[points.length - 1].x ?? x : x;
               this.y = points.length > 0 ? points[points.length - 1].y ?? y : y;
            }
         }
         return true;
      });
   }
   multiply(a: number | X2): XVector;
   multiply(x: number, y: number): XVector;
   multiply (a: number | X2, b = a as number) {
      if (typeof a === 'number') {
         return new XVector(this.x * a, this.y * b);
      } else {
         return this.multiply(a.x, a.y);
      }
   }
   round (base?: number): XVector {
      return base ? this.multiply(base).round().divide(base) : new XVector(Math.round(this.x), Math.round(this.y));
   }
   shift (angle: number, extent: number, origin = new XVector()) {
      if (extent === 0 && angle % 360 === 0) {
         return this.clone();
      } else {
         return origin.endpoint(this.angleFrom(origin) + angle, this.extentOf(origin) + extent);
      }
   }
   async step (speed: number, { x = this.x, y = this.y }) {
      await this.modulate((this.extentOf(x, y) / speed) * 1000, { x, y });
   }
   subtract(a: number | X2): XVector;
   subtract(x: number, y: number): XVector;
   subtract (a: number | X2, b = a as number) {
      if (typeof a === 'number') {
         return new XVector(this.x - a, this.y - b);
      } else {
         return this.subtract(a.x, a.y);
      }
   }
   value () {
      return { x: this.x, y: this.y };
   }
}

export class XText extends XAnchoredObject {
   private $text = {
      container: new PIXI.Container(),
      charset: '',
      content: '',
      dirty: false,
      plain: false,
      spacing: { x: 0, y: 0 } as X2,
      style: Object.assign({}, X.style),
      visible: false
   };
   /** The character set to use when measuring this text. */
   charset: string;
   /** The text content to render. */
   content: string;
   /** If true, this text will not handle special rendering codes. */
   plain: boolean;
   /** Additional spacing to apply to the text. */
   spacing = new XVector();
   constructor (properties: XTextProperties = {}) {
      super(properties);
      (({ charset = X.text.charset, content = '', plain = false, spacing = 0 }: XTextProperties = {}) => {
         this.charset = charset;
         this.content = content;
         this.plain = plain;
         if (typeof spacing === 'number') {
            this.spacing.inherit(spacing);
         } else {
            this.spacing.inherit(spacing.x ?? 0, spacing.y ?? 0);
         }
      })(properties);
   }
   compute () {
      const [ fontSize, fontName ] = this.$text.style.font.split('px ');
      const font = X.cache.fonts[fontName];
      if (font) {
         const fontScale = +fontSize / font.size;
         const lines = this.content.split('\n').map(section => {
            let total = 0;
            for (const char of section) {
               total += X.text.metrics(fontName, char, fontScale).x + this.spacing.x;
            }
            return total;
         });
         const metrics = X.text.metrics(fontName, this.charset, fontScale);
         return new XVector(Math.max(...lines), metrics.y + (metrics.y + this.spacing.y) * (lines.length - 1));
      } else {
         return new XVector();
      }
   }
   draw (style: XStyle) {
      if (this.container.alpha > 0) {
         if (!this.$text.visible) {
            this.$text.visible = true;
            this.container.addChildAt(this.$text.container, 0);
         }
         let update = false;
         if (this.charset !== this.$text.charset) {
            update = true;
            this.$text.charset = this.charset;
         }
         if (this.content !== this.$text.content) {
            update = true;
            this.$text.content = this.content;
         }
         if (this.plain !== this.$text.plain) {
            update = true;
            this.$text.plain = this.plain;
         }
         if (this.spacing.x !== this.$text.spacing.x) {
            update = true;
            this.$text.spacing.x = this.spacing.x;
         }
         if (this.spacing.y !== this.$text.spacing.y) {
            update = true;
            this.$text.spacing.y = this.spacing.y;
         }
         if (style.blend !== this.$text.style.blend) {
            update = true;
            this.$text.style.blend = style.blend;
         }
         if (style.border !== this.$text.style.border) {
            update = true;
            this.$text.style.border = style.border;
         }
         if (style.fill !== this.$text.style.fill) {
            update = true;
            this.$text.style.fill = style.fill;
         }
         if (style.font !== this.$text.style.font) {
            update = true;
            this.$text.style.font = style.font;
         }
         if (style.stroke !== this.$text.style.stroke) {
            update = true;
            this.$text.style.stroke = style.stroke;
         }
         if (style.tint !== this.$text.style.tint) {
            update = true;
            this.$text.style.tint = style.tint;
         }
         if (this.$text.dirty) {
            update = true;
            this.$text.dirty = false;
         }
         if (update) {
            let index = 0;
            const blend = X.blend(style.blend);
            const offset = { x: 0, y: 0 };
            const phase = X.timer.value / 1e3;
            const random = { x: 0, y: 0 };
            const state = Object.assign({}, style);
            const swirl = { p: 0, r: 0, s: 0 };
            const [ fontSize, fontName ] = style.font.split('px ');
            const font = X.cache.fonts[fontName];
            if (font) {
               const fontScale = +fontSize / font.size;
               const lines = this.content.split('\n').map(section => {
                  let total = 0;
                  for (const char of section) {
                     total += X.text.metrics(fontName, char, fontScale).x + this.spacing.x;
                  }
                  return total;
               });
               const metrics = X.text.metrics(fontName, this.charset, fontScale);
               const half = new XVector(
                  Math.max(...lines),
                  metrics.y + (metrics.y + this.spacing.y) * (lines.length - 1)
               ).divide(2);
               const base = new XVector(this.container.position.x, this.container.position.y).subtract(
                  half.add(half.multiply(this.anchor))
               );
               this.$text.container.removeChildren();
               while (index < this.content.length) {
                  const char = this.content[index++];
                  if (char === '\n') {
                     offset.x = 0;
                     offset.y += metrics.y + this.spacing.y;
                  } else if (!this.plain && char === '\xa7') {
                     const code = this.content.slice(index, this.content.indexOf('\xa7', index));
                     const [ key, value ] = code.split(':');
                     index += code.length + 1;
                     switch (key) {
                        case 'fill':
                           state.fill = value;
                           break;
                        case 'offset':
                           const [ offsetX, offsetY ] = value.split(',').map(value => +value);
                           offset.x = offsetX || 0;
                           offset.y = offsetY || 0;
                           break;
                        case 'random':
                           const [ randomX, randomY ] = value.split(',').map(value => +value);
                           random.x = randomX || 0;
                           random.y = randomY || 0;
                           break;
                        case 'stroke':
                           const strokeColor = X.cache.color.of(value);
                           state.border = strokeColor[3] > 0 ? state.border : 0;
                           state.stroke = value;
                           break;
                        case 'swirl':
                           const [ swirlR, swirlS, swirlP ] = value.split(',').map(value => +value);
                           swirl.r = swirlR || 0;
                           swirl.s = swirlS || 0;
                           swirl.p = swirlP || 0;
                           break;
                     }
                     if (random.x > 0 || random.y > 0 || (swirl.s > 0 && swirl.r > 0)) {
                        this.$text.dirty = true;
                     }
                  } else {
                     let x = base.x - state.border / 2 + offset.x;
                     let y = base.y - state.border / 2 + offset.y;
                     if (random.x > 0) {
                        x += random.x * (Math.random() - 0.5);
                     }
                     if (random.y > 0) {
                        y += random.y * (Math.random() - 0.5);
                     }
                     if (swirl.s > 0 && swirl.r > 0) {
                        const endpoint = new XVector(x, y).endpoint(
                           ((phase * 360 * swirl.s) % 360) + index * (360 / swirl.p),
                           swirl.r
                        );
                        x = endpoint.x;
                        y = endpoint.y;
                     }
                     if (font) {
                        const info = font.glyphs[char.charCodeAt(0)];
                        if (info) {
                           const fill = X.cache.color.of(state.fill);
                           const sprite = PIXI.Sprite.from(info.image.value);
                           sprite.alpha = fill[3];
                           sprite.position.set(
                              x + info.metrics.x - this.container.position.x + font.shift.x,
                              y + info.metrics.y - this.container.position.y + font.shift.y
                           );
                           sprite.scale.set(fontScale);
                           sprite.blendMode = blend;
                           sprite.tint = X.color2hex(fill);
                           this.$text.container.addChild(sprite);
                        }
                        offset.x += X.text.metrics(fontName, char, fontScale).x + this.spacing.x;
                     }
                  }
               }
            } else {
               this.$text.dirty = true;
            }
         }
      } else if (this.$text.visible) {
         this.$text.visible = false;
         this.container.removeChildAt(0);
      }
   }
}

export const X = {
   /** Retrieve an audio buffer from the cache. */
   audio (source: string) {
      return X.cache.audio.of(source);
   },
   /** Returns an asset for a given audio source. */
   audioAsset (
      /** The audio's source. */
      source: string,
      {
         /**
          * The extra duration in which to keep this asset's source audio in memory after this asset and all of its
          * siblings (other assets which share this asset's source audio) are unloaded.
          */
         cache = 0,
         /** The data modifier to apply to the audio. */
         transformer = void 0 as ((value: number, index: XVector, total: X2) => number) | void,
         /** The trim to apply to the audio. */
         trim: { start = 0, stop = 0 } = {}
      } = {}
   ): XAudio {
      const asset = new XAsset({
         async loader () {
            const assets = (X.cache.audioAssets[source] ??= []);
            assets.includes(asset) || assets.push(asset);
            const audio = await X.cache.audio.of(source);
            if (start ?? stop ?? transformer) {
               const c = audio.numberOfChannels;
               const b = Math.round(audio.sampleRate * ((start < 0 ? audio.duration : 0) + start));
               const l = Math.round(audio.sampleRate * ((stop < 0 ? 0 : audio.duration) - stop)) - b;
               const index = new XVector(-1);
               const clone = new AudioBuffer({ length: l, numberOfChannels: c, sampleRate: audio.sampleRate });
               while (++index.y < c) {
                  const data = audio.getChannelData(index.y).slice(b, b + l);
                  if (transformer) {
                     const total = { x: c, y: l };
                     while (++index.x < l) {
                        data[index.x] = transformer(data[index.x], index, total);
                     }
                     index.x = -1;
                  }
                  clone.copyToChannel(data, index.y);
               }
               return clone;
            } else {
               return audio;
            }
         },
         source,
         async unloader () {
            const assets = (X.cache.audioAssets[source] ??= []);
            X.pause(cache).then(() => {
               assets.includes(asset) && assets.splice(assets.indexOf(asset), 1);
               if (assets.length === 0) {
                  X.cache.audio.delete(source);
               }
            });
         }
      });
      return asset;
   },
   /** Gets the blend mode constant for the given composite operation. */
   blend (input: GlobalCompositeOperation) {
      switch (input) {
         case 'lighten':
            return PIXI.BLEND_MODES.ADD;
         case 'multiply':
            return PIXI.BLEND_MODES.MULTIPLY;
         case 'screen':
            return PIXI.BLEND_MODES.SCREEN;
         default:
            return PIXI.BLEND_MODES.NORMAL;
      }
   },
   /** Caches for various engine systems. Internal use only. */
   cache: {
      audio: new XCache<string, Promise<AudioBuffer>>(
         source =>
            new Promise<AudioBuffer>(resolve => {
               const request = Object.assign(new XMLHttpRequest(), { responseType: 'arraybuffer' });
               request.addEventListener('load', () => X.context.decodeAudioData(request.response, resolve));
               request.open('GET', source, true);
               request.send();
            })
      ),
      audioAssets: {} as XKeyed<XAudio[]>,
      chalkboard: null as PIXI.Extract | null,
      checks: [] as { condition: () => boolean; resolve: () => void }[],
      color: new XCache<string, XColor>(input => {
         const element = document.createElement('x');
         element.style.color = input;
         document.body.appendChild(element);
         const color = getComputedStyle(element)
            .color.split('(')[1]
            .slice(0, -1)
            .split(', ')
            .map(value => +value);
         element.remove();
         color.length === 3 && color.push(1);
         return color as XColor;
      }),
      data: new XCache<string, Promise<XBasic>>(source => fetch(source).then(value => value.json() as Promise<XBasic>)),
      dataAssets: {} as XKeyed<XData[]>,
      fonts: {} as XKeyed<XFont>,
      image: new XCache<string, Promise<PIXI.BaseTexture>>(async source => {
         const texture = PIXI.BaseTexture.from(source);
         await texture.resource.load();
         return texture;
      }),
      imageAssets: {} as XKeyed<XImage[]>,
      sprites: new XCache<XImage, Set<PIXI.Sprite>>(() => new Set()),
      tasks: new Map<any, () => void>(),
      texture: null as PIXI.Texture | null
   },
   /** It's complicated. */
   chain<A, B> (input: A, handler: (input: A, loop: (input: A) => B) => B) {
      const loop = (input: A) => handler(input, loop);
      return loop(input);
   },
   /** For pixel extraction. Internal use only. */
   get chalkboard () {
      return (X.cache.chalkboard ??= new PIXI.Extract(new PIXI.Renderer({ antialias: false, backgroundAlpha: 0 })));
   },
   /** Retrieve a color value from the cache. */
   color (source: string) {
      return X.cache.color.of(source);
   },
   /** Converts an `XColor` instance into its integer representation. */
   color2hex (color: XColor, alpha = false) {
      return parseInt(
         color
            .slice(0, alpha ? 4 : 3)
            .map(value => Math.round(value).toString(16).padStart(2, '0'))
            .join(''),
         16
      );
   },
   /** For audio decoding. Internal use only. */
   context: new AudioContext(),
   /** Returns the area of an `XCrop` instance applied to an image. */
   crop (source: { width: number; height: number }, crop: XCrop): XArea {
      const x = (crop.left < 0 ? source.width : 0) + crop.left;
      const y = (crop.top < 0 ? source.height : 0) + crop.top;
      return {
         h: (crop.bottom < 0 ? 0 : source.height) - crop.bottom - y,
         w: (crop.right < 0 ? 0 : source.width) - crop.right - x,
         x,
         y
      };
   },
   /** Retrieve a data value from the cache. */
   data<A> (source: string) {
      return X.cache.data.of<Promise<A>>(source);
   },
   /** Returns an asset for a given data source. */
   dataAsset<A extends XBasic = XBasic> (
      source: string,
      { cache = 0, modifier = void 0 as ((data: XBasic) => A) | void } = {}
   ): XData<A> {
      const asset = new XAsset<A>({
         async loader () {
            const assets = (X.cache.dataAssets[source] ??= []);
            assets.includes(asset) || assets.push(asset);
            const data = await X.cache.data.of(source);
            if (modifier) {
               return modifier(data);
            } else {
               return data as A;
            }
         },
         source,
         async unloader () {
            const assets = (X.cache.dataAssets[source] ??= []);
            X.pause(cache).then(() => {
               assets.includes(asset) && assets.splice(assets.indexOf(asset), 1);
               if (assets.length === 0) {
                  X.cache.data.delete(source);
               }
            });
         }
      });
      return asset;
   },
   /** Allows a promise to be resolved outside of its callback. */
   hyperpromise<A = void> () {
      let hyperresolve: (value: A | PromiseLike<A>) => void;
      const promise = new Promise<A>(resolve => {
         hyperresolve = resolve;
      });
      return { promise, resolve: hyperresolve! };
   },
   /** Retrieve a base texture from the cache. */
   image (source: string) {
      return X.cache.image.of(source);
   },
   /** Returns an asset for a given image source. */
   imageAsset (source: string, { cache = 0, shader = void 0 as XShader | void } = {}): XImage {
      const asset = new XAsset({
         async loader () {
            const assets = (X.cache.imageAssets[source] ??= []);
            assets.includes(asset) || assets.push(asset);
            let image = await X.cache.image.of(source);
            if (image.width === 0 || image.height === 0) {
               image = PIXI.BaseTexture.from(await createImageBitmap(new ImageData(1, 1)));
            } else if (shader) {
               const sprite = PIXI.Sprite.from(image);
               const data = new Uint8ClampedArray(X.chalkboard.pixels(sprite));
               sprite.destroy();
               const x4 = image.width * 4;
               const index = { x: -1, y: -1 };
               const total = { x: image.width, y: image.height };
               while (++index.x < image.width) {
                  const n4 = index.x * 4;
                  while (++index.y < image.height) {
                     let step = index.y * x4 + n4;
                     const color = shader([ data[step++], data[step++], data[step++], data[step++] ], index, total);
                     data[--step] = color[3];
                     data[--step] = color[2];
                     data[--step] = color[1];
                     data[--step] = color[0];
                  }
                  index.y = -1;
               }
               image = PIXI.BaseTexture.from(await createImageBitmap(new ImageData(data, image.width)));
            }
            const texture = PIXI.Texture.from(image);
            if (X.cache.sprites.has(asset)) {
               for (const sprite of X.cache.sprites.get(asset)!) {
                  sprite.texture = texture;
               }
            }
            return texture;
         },
         source,
         async unloader () {
            const value = asset.value;
            const assets = (X.cache.imageAssets[source] ??= []);
            X.pause(cache).then(() => {
               assets.includes(asset) && assets.splice(assets.indexOf(asset), 1);
               if (assets.length === 0) {
                  if (X.cache.sprites.has(asset)) {
                     for (const sprite of X.cache.sprites.get(asset)!) {
                        sprite.texture = X.texture;
                     }
                  }
                  X.cache.image.delete(source);
                  value.destroy(true);
               }
            });
         }
      });
      return asset;
   },
   /** Creates an inventory of assets. This inventory is a superset of `XAsset` and will cause contained assets to be retained when this inventory is loaded. */
   inventory<A extends XAsset[]> (...assets: A): XInventory<A> {
      const inventory = new XAsset({
         async loader () {
            return await Promise.all(assets.map(asset => asset.load(inventory))).then(() => assets);
         },
         source: assets.map(asset => asset.source).join('\n'),
         async unloader () {
            await Promise.all(assets.map(asset => asset.unload(inventory)));
         }
      }) as XInventory<A>;
      return inventory;
   },
   /** Various math tools for common game development needs. */
   math: {
      bezier (value: number, ...points: number[]): number {
         if (points.length === 0) {
            return value;
         } else if (points.length === 1) {
            return points[0];
         } else {
            return X.math.bezier(
               value,
               ...points.slice(0, -1).map((point, index) => point * (1 - value) + points[index + 1] * value)
            );
         }
      },
      cardinal (angle: number): XCardinal {
         const trueAngle = angle % 360;
         if (trueAngle < -315 || trueAngle > 315) {
            return 'right';
         } else if (trueAngle <= -225) {
            return 'down';
         } else if (trueAngle < -135) {
            return 'left';
         } else if (trueAngle <= -45) {
            return 'up';
         } else if (trueAngle < 45) {
            return 'right';
         } else if (trueAngle <= 135) {
            return 'down';
         } else if (trueAngle < 225) {
            return 'left';
         } else {
            return 'up';
         }
      },
      format (value: number) {
         return Math.round(value);
      },
      intersection (a1: X2, a2: X2, b1: X2, b2: X2) {
         return (
            X.math.rotation(a1, b1, b2) !== X.math.rotation(a2, b1, b2) &&
            X.math.rotation(a1, a2, b1) !== X.math.rotation(a1, a2, b2)
         );
      },
      linear (value: number, ...points: number[]) {
         if (points.length === 0) {
            return value;
         } else if (points.length === 1) {
            return points[0];
         } else if (value <= 0) {
            return X.math.remap(value, points[0], points[1]);
         } else if (1 <= value) {
            return X.math.remap(value, points[points.length - 2], points[points.length - 1]);
         } else {
            const supervalue = value * (points.length - 1);
            const index = Math.floor(supervalue);
            return X.math.remap(supervalue % 1, points[index], points[index + 1]);
         }
      },
      remap (value: number, min2: number, max2: number, min1 = 0, max1 = 1) {
         return ((value - min1) * (max2 - min2)) / (max1 - min1) + min2;
      },
      rotation (a1: X2, a2: X2, a3: X2) {
         return (a3.y - a1.y) * (a2.x - a1.x) > (a2.y - a1.y) * (a3.x - a1.x);
      },
      wave (value: number) {
         return Math.sin(((value + 0.5) * 2 - 1) * Math.PI) / 2 + 0.5;
      },
      weigh<A extends string> (input: XProvider<[A, number][]>, modifier = Math.random()) {
         const weights = X.provide(input);
         let total = 0;
         for (const entry of weights) {
            total += entry[1];
         }
         const value = modifier * total;
         for (const entry of weights) {
            if (value > (total -= entry[1])) {
               return entry[0];
            }
         }
      }
   },
   /** Parses JSON data previously stringified with `X.stringify`. */
   parse (text: string) {
      return JSON.parse(text, (key, value) => {
         if (value === '\x00') {
            return Infinity;
         } else if (value === '\x01') {
            return -Infinity;
         } else if (value === '\x02') {
            return NaN;
         } else {
            return value;
         }
      });
   },
   /** Returns a promise which resolves after the given duration in milliseconds. */
   pause (duration = 0) {
      if (duration === Infinity) {
         return new Promise<void>(() => {});
      } else {
         const time = X.timer.value;
         return X.when(() => duration <= X.timer.value - time);
      }
   },
   /** Populates an array with values from the given provider. */
   populate: ((size: number, provider: any) => {
      let index = 0;
      const array: any[] = [];
      while (index < size) {
         array.push(X.provide(provider, index++));
      }
      return array;
   }) as {
      <A extends (index: number) => unknown>(size: number, provider: A): ReturnType<A>[];
      <A>(size: number, provider: A): A[];
   },
   /** Returns the value of a provider. */
   provide<A extends XProvider<unknown, unknown[]>> (
      provider: A,
      ...args: A extends XProvider<any, infer B> ? B : never
   ): A extends XProvider<infer B, any[]> ? B : never {
      return typeof provider === 'function' ? X.provide(provider(...args)) : provider;
   },
   /** Tools for creating sound effects. */
   sound: {
      convolver (context: AudioContext, duration: number, ...pattern: number[]) {
         const convolver = context.createConvolver();
         convolver.buffer = X.sound.impulse(context, duration, ...pattern);
         return convolver;
      },
      delay (context: AudioContext, offset: number, feedback: number, echoes = Infinity) {
         const gain = context.createGain();
         gain.gain.value = feedback;
         const delay = context.createDelay(Math.min(offset * echoes, 60));
         delay.delayTime.value = offset;
         gain.connect(delay);
         delay.connect(gain);
         return gain;
      },
      filter (context: AudioContext, type: BiquadFilterType, frequency: number) {
         const filter = context.createBiquadFilter();
         filter.type = type;
         filter.frequency.value = frequency;
         return filter;
      },
      impulse (context: AudioContext, duration: number, ...pattern: number[]) {
         let index = -1;
         const size = context.sampleRate * duration;
         const buffer = context.createBuffer(2, size, context.sampleRate);
         while (++index < size) {
            let channel = 0;
            while (channel < buffer.numberOfChannels) {
               try {
                  const data = buffer.getChannelData(channel++);
                  data[index] = (Math.random() * 2 - 1) * X.math.bezier(index / size, ...pattern);
                  data[index] = (Math.random() * 2 - 1) * X.math.bezier(index / size, ...pattern);
               } catch (error) {}
            }
         }
         return buffer;
      }
   },
   /** Master speed value. This impacts all calculations in the engine. */
   speed: new XNumber(1),
   /** Prints styled status messages to the console. */
   status (
      text: string,
      {
         backgroundColor = '#000',
         color = '#fff',
         fontFamily = 'Courier New',
         fontSize = '16px',
         padding = '4px 8px'
      } = {}
   ) {
      console.log(
         `%c${text}`,
         `background-color:${backgroundColor};color:${color};font-family:${fontFamily};font-size:${fontSize};padding:${padding};`
      );
   },
   /** Serializes an object to JSON, with support for `Infinity` and `NaN` values. */
   stringify (value: any) {
      return JSON.stringify(value, (key, value) => {
         if (value === Infinity) {
            return '\x00';
         } else if (value === -Infinity) {
            return '\x01';
         } else if (typeof value === 'number' && value !== value) {
            return '\x02';
         } else {
            return value;
         }
      });
   },
   /** The default style for all objects. */
   style: {
      blend: 'source-over',
      border: 1,
      fill: 'transparent',
      font: '10px monospace',
      stroke: 'transparent',
      tint: 16777215
   } as XStyle,
   /** Creates an image bitmap from the given pixel data. */
   synthesize (colors: XColor[][]) {
      return createImageBitmap(new ImageData(new Uint8ClampedArray(colors.flat(2)), colors.length));
   },
   /** Functions for handling text. */
   text: {
      /** The default measurement character set for `XText` objects. */
      charset: '/0123456789=ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz',
      /** Functions for font creation and management. */
      font: {
         /** Creates a font from the given options. */
         async create (
            /** The font family. */
            name: string,
            /** The font size to create. Note that when fonts are drawn, they will scale based on the ratio of this value and the drawn font size. */
            size: number,
            /** The offset to move each glyph by when the font is drawn. */
            shift: X2,
            /** The character set to generate. */
            chars: string,
            /** The resolution to generate this font at. */
            resolution: number
         ) {
            const style = new PIXI.TextStyle({ fontFamily: name, fontSize: size, fill: 'white', padding: 20 });
            const font = PIXI.BitmapFont.from(name, style, {
               chars,
               resolution,
               textureHeight: resolution * 100,
               textureWidth: resolution * 100
            });
            const extract = X.chalkboard;
            const value: XFontData = {
               name,
               glyphs: await Promise.all(
                  Object.entries(font.chars).map(async ([ code, char ]) => {
                     const sprite = PIXI.Sprite.from(char.texture);
                     const submetrics = PIXI.TextMetrics.measureText(String.fromCharCode(+code), style);
                     const value: XGlyphData = {
                        code,
                        margin: char.xAdvance ?? 0,
                        metrics: {
                           x: char.xOffset ?? 0,
                           y: char.yOffset ?? 0,
                           w: submetrics.width,
                           h: submetrics.height
                        },
                        source: await extract.base64(sprite, 'image/png', 1)
                     };
                     sprite.destroy();
                     return value;
                  })
               ),
               shift,
               size
            };
            font.destroy();
            return value;
         },
         /** Import and load font data, to prepare it for registry. */
         async import ({ name, glyphs, shift, size }: XFontData) {
            return {
               name,
               font: {
                  glyphs: Object.fromEntries(
                     await Promise.all(
                        glyphs.map(async ({ code, margin, metrics, source }) => {
                           const image = X.imageAsset(source);
                           await image.load();
                           return [ code, { image, margin, metrics } ] as [string, XGlyph];
                        })
                     )
                  ),
                  shift,
                  size
               }
            };
         },
         /** Preload fonts by first rendering them on the DOM. */
         async preload (
            /** The font families to preload. */
            fonts: string[],
            /** The duration in milliseconds to wait. The longer this duration, the more likely fonts are to preload. The default value of `300` is well-tested, but may not work on all machines. Try to increase this value if the fonts do not load correctly. */
            time = 300
         ) {
            const elements = fonts.map(name => {
               const element = document.createElement('div');
               element.style.fontFamily = name;
               element.style.opacity = '0';
               element.textContent = 'x';
               document.body.appendChild(element);
               return element;
            });
            await X.pause(time);
            for (const element of elements) {
               element.remove();
            }
         }
      },
      /** Get the metrics for a given font and content. */
      metrics (name: string, content: string, scale: number): X2 {
         if (name in X.cache.fonts) {
            let x = 0;
            let y = 0;
            let index = 0;
            const font = X.cache.fonts[name];
            while (index < content.length) {
               const info = font.glyphs[content.charCodeAt(index++)];
               if (info) {
                  x += (info.metrics.x + info.metrics.w + info.margin) / 2;
                  y = Math.max(y, info.metrics.y + info.metrics.h);
               }
            }
            return new XVector(x, y).multiply(scale);
         } else {
            return new XVector();
         }
      },
      /** Register a font, making it available for use in the engine. */
      register (name: string, font: XFont) {
         X.cache.fonts[name] = font;
      },
      /** Unregister a font, taking it out of availability for use in the engine. */
      unregister (name: string) {
         delete X.cache.fonts[name];
      }
   },
   /** A placeholder texture. Internal use only. */
   get texture () {
      return (X.cache.texture ??= new PIXI.Texture(new PIXI.BaseTexture()));
   },
   /** The timing system used throughout the engine. */
   timer: {
      /** The timer's host. */
      host: new XHost<{ tick: [number] }>().on('tick', () => {
         for (const check of X.cache.checks) {
            check.condition() && (void X.cache.checks.splice(X.cache.checks.indexOf(check), 1), check.resolve());
         }
      }),
      /** Initializes the timer. */
      init () {
         let pre = performance.now();
         return setInterval(() => {
            const now = performance.now();
            const delta = (now - pre) * X.speed.value;
            X.timer.value += delta;
            X.timer.host.fire('tick', delta);
            pre = now;
         });
      },
      /** The timer's current time value. */
      value: 0
   },
   /** Returns a promise which resolves when the function returns true. */
   when (condition: () => boolean) {
      const { promise, resolve } = X.hyperpromise();
      X.cache.checks.push({ condition, resolve });
      return promise;
   }
};

AudioParam.prototype.interpolate = function (duration: number, ...points: number[]) {
   return XNumber.prototype.interpolate.call(this, duration, ...points);
};

AudioParam.prototype.modulate = function (duration: number, ...points: number[]) {
   return XNumber.prototype.modulate.call(this, duration, ...points);
};

PIXI.settings.RESOLUTION = 1;
PIXI.settings.ROUND_PIXELS = false;
PIXI.settings.SCALE_MODE = PIXI.SCALE_MODES.NEAREST;
